# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019-2020 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Functions to load and work with raw numeric data."""

import functools
import inspect
import math
import re
import sys
from fractions import Fraction

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..compat import MAJOR, long

_API_START = api_start(globals())
# -------------------------------


#: Maximum entrywise absolute deviation allowed for numeric equality checks.
TOLERANCE = 1e-6


def load_shape(shape, squareMatrix=False, wildcards=False):
    """Parse the argument as a matrix shape.

    PICOS uses this function whenever you supply a shape parameter to a method.

    A scalar argument is treated as the length of a column-vector. If the shape
    contains :obj:`None`, it is treated as a wildcard (any dimensionality).

    :param bool squareMatrix: If :obj:`True`, a scalar argument is treated as
        the side/diagonal length of a square matrix, and any other argument is
        validated to be square. If :obj:`False`, a scalar argument is treated
        as the length of a column vector.
    :param bool wildcards: Whether the wildcard token :obj:`None` is allowed.
    """
    if shape is None:
        shape = (None, None)
    elif isinstance(shape, int):
        if squareMatrix:
            shape = (shape, shape)
        else:
            shape = (shape, 1)
    elif not isinstance(shape, tuple) and not isinstance(shape, list):
        raise TypeError("Shapes must be given as None, int, tuple or list.")
    elif len(shape) == 1:
        shape = (shape[0], 1)
    elif len(shape) == 0:
        shape = (1, 1)
    elif len(shape) != 2:
        raise TypeError("Shapes must be two-dimensional.")

    shape = (
        None if shape[0] is None else int(shape[0]),
        None if shape[1] is None else int(shape[1]))

    if not wildcards and None in shape:
        raise ValueError("Invalid shape (wildcards not allowed): {}."
            .format(glyphs.shape(shape)))

    if 0 in shape:
        raise ValueError("Invalid shape (zero-dimensional axis): {}."
            .format(glyphs.shape(shape)))

    if squareMatrix and shape[0] != shape[1]:
        raise ValueError("Invalid shape for a square matrix: {}."
            .format(glyphs.shape(shape)))

    return shape


def blend_shapes(baseShape, defaultShape):
    """Replace wildcards in one shape with entries of the other.

    :param baseShape: Primary shape, usually with wildcards.
    :type baseShape: tuple(int or None)
    :param defaultShape: Secondary shape with fallback entries.
    :type defaultShape: tuple(int or None)
    """
    return (
        defaultShape[0] if baseShape[0] is None else baseShape[0],
        defaultShape[1] if baseShape[1] is None else baseShape[1])


def should_be_sparse(shape, numNonZero):
    """Decide whether a matrix is considered sparse.

    :param tuple(int) shape: The shape of the matrix in question.
    :param int numNonZero: Number of non-zero elements of the matrix.
    """
    n, m = shape
    l = (n*m)**0.5
    return numNonZero < l * math.log(l)


def load_data(value, shape=None, typecode=None, sparse=None,
        alwaysCopy=True, legacy=False):
    r"""Load a constant numeric data value as a CVXOPT (sparse) matrix.

    As a user, you never need to call this manually, but you should be aware
    that PICOS uses this function on any raw data you supply as an operand when
    working with PICOS expressions. For instance, you can just add a NumPy
    matrix or an algebraic string such as ``"I"`` to such an expression without
    worrying about any conversion.

    :Supported data types:

        - A NumPy matrix: :class:`numpy.ndarray` or :class:`numpy.matrix`.
        - A CVXOPT matrix: :obj:`cvxopt.matrix` or :obj:`cvxopt.spmatrix`.
        - A constant PICOS expression: :class:`~.exp_affine.AffineExpression` or
          :class:`~.exp_affine.ComplexAffineExpression`.
        - A Python scalar: :class:`int`, :class:`float` or :class:`complex`.
        - A flat :class:`tuple` or :class:`list` containing scalars or a
          :class:`range`, all representing a column vector.
        - A nested :class:`tuple` or :class:`list` containing scalars. The outer
          level represents rows and the inner level represents the rows'
          entries. Allows you to define a :math:`2 \times 3` matrix like this:

          .. code-block:: python

              A = [[1, 2, 3],
                   [4, 5, 6]]

        - A verbatim string description, with rows separated by newline and
          columns separated by whitespace. The same :math:`2 \times 3` matrix as
          above:

          .. code-block:: python

              A = '''1 2 3
                     4 5 6'''

        - An algebraic string description:

          .. list-table::
            :widths: 1 99

            * - ``"|a|"``
              - A matrix with all entries equal to :math:`a`.
            * - ``"|a|(m,n)"``
              - A :math:`m \times n` matrix with all entries equal to :math:`a`.
            * - ``"e_i,j(m,n)"``
              - A :math:`m \times n` matrix with a :math:`1` at :math:`(i,j)`,
                indexed from :math:`(1,1)`` to :math:`(m,n)``.
            * - ``"e_i(m,n)"``
              - A :math:`m \times n` matrix with a single :math:`1` on the
                :math:`i`-th coordinate, indexed from :math:`1` in column-major
                order.
            * - ``"I"``
              - The identity matrix.
            * - ``"I(n)"``
              - The :math:`n \times n` identiy matrix.
            * - ``"a…"``
              - The matrix given by ``"…"`` but multiplied by :math:`a`.

        Different matrix operations such as addition or multiplication have
        different requirements with respect to the operands' shapes. The shape
        of any PICOS expression involved will always be maintained. But when an
        operation involves both a PICOS expression and raw data, then PICOS will
        try to broadcast or reshape the raw data such that the operation can be
        performed.

    :Broadcasting and reshaping rules:

        - An input vector without a second axis (for instance a non-nested
          :class:`tuple` or :class:`list` or a :class:`range`) is interpreted as
          a row vector if the target shape is of the form ``(None, n)`` with
          :math:`n > 1`, otherwise it is interpreted as a column vector.
        - If the target shape is :obj:`None` or ``(None, None)``, then the
          input's shape is maintained.
        - If the target shape contains :obj:`None` exactly once, that occurence
          is replaced by the respective dimensionality of the input data shape.
        - A scalar is copied (broadcasted) to fill the target shape.
        - A column (row) vector is copied horizontally (vertically) if its
          length matches the target row (column) count.
        - Reshaping from matrix to vector: A matrix is vectorized in
          column-major (row-major) order if the target shape is a column (row)
          vector whose length matches the number of matrix entries.
        - Reshaping from vector to matrix: A column (row) vector is treated as
          the column-major (row-major) vectorization of the target matrix if the
          former's length matches the number of the latter's entries.
        - All other combinations of input and target shape raise an exception.
        - When an algebraic string description specifies no shape, then the
          shape argument must be supplied. When both the string and the shape
          argument specify a shape, then they must be consistent (no
          broadcasting or reshaping is applied in this case).

    :param shape: The shape of the resulting matrix. If the input data is of
        another shape, broadcasting or reshaping is used if possible, otherwise
        an exception is raised. An integer is treated as the length of a column
        vector. If this is :obj:`None`, then the target's shape is the input's.
        If only the target number of rows or columns is :obj:`None`, then only
        that quantity is chosen according to the input.
    :type shape: int or tuple or list or None

    :param str typecode: The numeric type of the resulting matrix. Either
        ``'d'`` (float), ``'z'`` (complex) or ``'i'`` (integer). If the input
        data is not already of this type, then it will be converted if possible.
        If this is not possible, then an exception is raised. If this is
        :obj:`None`, then the output type is chosen based on the input data.

    :param sparse: If :obj:`True`, a sparse matrix is returned. If :obj:`False`,
        a dense matrix is returned. If :obj:`None`, it depends on the sparsity
        pattern of the input data. If the typecode argument is ``'i'``, then
        a value of :obj:`True` is not allowed and the returned matrix is dense.
    :type sparse: bool or None

    :param bool alwaysCopy: If :obj:`True`, then a copy of the input data is
        returned even if it already equals the output data. If :obj:`False`,
        the input value can be returned as such if it is already a CVXOPT matrix
        of the desired shape, typecode, and sparsity.

    :param bool legacy: Be compatible with the old ``retrieve_matrix`` function.
        In particular, if the target shape contains :obj:`None` exactly once and
        the input data is scalar, treat this as a matrix multiplication case and
        return the scalar times an identity matrix of appropriate size.

    :returns: A :class:`tuple` whose first entry is the loaded matrix and whose
        second argument is a short string for representing the data within
        algebraic expression strings.

    :Example:

    >>> from picos.expressions.data import load_data
    >>> # Data as (nested) list:
    >>> load_data([1,2,3])
    (<3x1 matrix, tc='i'>, '[3×1]')
    >>> load_data([[1,2,3]])
    (<1x3 matrix, tc='i'>, '[1×3]')
    >>> A = [[1,2,3],
    ...      [4,5,6]]
    >>> load_data(A)
    (<2x3 matrix, tc='i'>, '[2×3]')
    >>> # Data as string:
    >>> value, string = load_data('e_14(7,2)')
    >>> print(string)
    [7×2:e_7,2]
    >>> print(value) #doctest: +NORMALIZE_WHITESPACE
    [   0        0       ]
    [   0        0       ]
    [   0        0       ]
    [   0        0       ]
    [   0        0       ]
    [   0        0       ]
    [   0        1.00e+00]
    >>> load_data('5.3I', (2,2))
    (<2x2 sparse matrix, tc='d', nnz=2>, '5.3·I')
    """
    from .expression import Expression
    from .exp_affine import ComplexAffineExpression

    def load_sparse(V, I, J, shape, typecode):
        """Create a CVXOPT sparse matrix."""
        # HACK: Work around CVXOPT not supporting integer sparse matrices:
        #       Create a real sparse matrix for now (will be converted later).
        #       Note that users may not request both sparsity and integrality.
        typecode = "d" if typecode == "i" else typecode

        try:
            if typecode:
                return cvxopt.spmatrix(V, I, J, shape, typecode)
            else:
                return cvxopt.spmatrix(V, I, J, shape)
        except TypeError as error:
            # Attempt to convert complex typed but real valued input to a real
            # typed output matrix.
            if typecode == "d":
                realV = [x.real for x in V]
                if realV == V:
                    try:
                        return cvxopt.spmatrix(realV, I, J, shape, typecode)
                    except Exception:
                        pass

            raise TypeError("Failed to create a CVXOPT sparse matrix of shape "
                "{} and type '{}': {}.".format(glyphs.shape(shape), typecode,
                str(error).capitalize()))

    def load_dense(value, shape, typecode):
        """Create a CVXOPT dense matrix."""
        try:
            if typecode:
                return cvxopt.matrix(value, shape, typecode)
            else:
                return cvxopt.matrix(value, shape)
        except TypeError as error:
            # Attempt to convert complex (real) typed but real/integer (integer)
            # valued input to a real/integer (integer) typed output matrix.
            if typecode in "id":
                try:
                    complexValue = cvxopt.matrix(value, shape, "z")
                except Exception:
                    pass
                else:
                    if not any(complexValue.imag()):
                        if typecode == "d":
                            return complexValue.real()
                        else:
                            realData = list(complexValue.real())
                            intData  = [int(x) for x in realData]
                            if intData == realData:
                                try:
                                    return cvxopt.matrix(intData, shape, "i")
                                except Exception:
                                    pass

            raise TypeError("Failed to create a CVXOPT dense matrix of shape "
                "{} and type '{}': {}.".format(glyphs.shape(shape), typecode,
                str(error).capitalize()))

    def simple_vector_as_row(shape):
        """Whether a single-axis vector should be a row or column vector."""
        return shape[0] is None and shape[1] is not None and shape[1] > 1

    def broadcast_error():
        """Raise a broadcasting related :class:TypeError."""
        raise TypeError("Cannot broadcast or reshape from {} to {}{}."
            .format(glyphs.shape(inShape), glyphs.shape(shape), " read as {}"
                .format(glyphs.shape(outShape)) if shape != outShape else ""))

    def scalar(x, typecode=None):
        """Load a scalar as either a complex, float, or int."""
        x = complex(x)

        if typecode == "z":
            return x  # We are content with complex.
        else:
            if not x.imag:
                x = x.real

                if typecode == "d":
                    return x  # We are content with real.
                else:
                    if int(x) == x:
                        return int(x)
                    elif not typecode:
                        return x  # We are not strict and real is the best.
            elif not typecode:
                return x  # We are not strict and complex is the best.

        raise TypeError(
            "Cannot cast {} according to typecode '{}'.".format(x, typecode))

    # Normalize the shape argument to a two-dimensional tuple of int or None,
    # where None means any dimensionality.
    shape = load_shape(shape, wildcards=True)

    # Validate the typecode argument.
    if typecode is not None and typecode not in "dzi":
        raise ValueError("Typecode argument not a valid CVXOPT typecode: {}."
            .format(typecode))

    # Validate the sparsity argument.
    if sparse not in (None, True, False):
        raise ValueError("Sparsity argument must be True, False or None.")

    # CVXOPT sparse matrices may not be integer typed.
    if typecode == "i":
        if sparse:
            raise TypeError("Sparse integer matrices are not implemented with "
                "CVXOPT, which PICOS uses as its matrix backed.")
        else:
            sparse = False

    # Conversions must retrieve the input shape for further processing.
    inShape = None

    # Allow conversions to specify their own string description.
    string = None

    # Allow Python 3 range types.
    if MAJOR >= 3 and isinstance(value, range):
        value = list(value)

    # Try to refine a PICOS expression to a constant affine expression.
    if isinstance(value, Expression):
        value = value.refined

    # Convert the data to a CVXOPT (sparse) matrix of proper shape and type.
    if isinstance(value, (int, long, float, complex, numpy.number)):
        # Convert NumPy numbers to Python ones.
        if isinstance(value, numpy.number):
            if isinstance(value, numpy.integer):
                value = int(value)  # Result can be a Python 2 long.
            elif isinstance(value, numpy.floating):
                value = float(value)
            elif isinstance(value, numpy.complexfloating):
                value = complex(value)
            else:
                assert False, "Unexpected NumPy numeric type {}.".format(
                    type(value).__name__)

        # CVXOPT is limited by the system's word length.
        if isinstance(value, (int, long)):
            if value > sys.maxsize or value < -sys.maxsize - 1:
                raise ValueError(
                    "The number {} is too large to be loaded by PICOS."
                    .format(value))

            if isinstance(value, long):
                value = int(value)
                assert isinstance(value, int), "Failed to convert from long to"\
                    " int despite being smaller than sys.maxsize."

        inShape  = (1, 1)
        outShape = blend_shapes(shape, inShape)

        string = glyphs.scalar(value)
        if value and outShape != (1, 1):  # Don't use the matrix glyph on zero.
            string = glyphs.matrix(string)

        if not value and sparse is not False:
            value = load_sparse([], [], [], outShape, typecode)
        else:
            value = load_dense(value, outShape, typecode)

            if sparse:
                value = cvxopt.sparse(value)
    elif isinstance(value, cvxopt.matrix):
        # NOTE: Since the input is already a CVXOPT data type, it is possible
        #       that it can be returned without modification. The 'alwaysCopy'
        #       parameter, if set to True, prevents this. As soon as a copy is
        #       made during processing of the input we set 'alwaysCopy' to False
        #       to prevent an unnecessary second copy of the processed input to
        #       be made later on.

        # If sparse is requested, let CVXOPT handle the transformation first.
        if sparse:
            value = cvxopt.sparse(value)
            return load_data(value, shape, typecode, sparse, False, legacy)

        # Refine the output shape.
        inShape  = value.size
        outShape = blend_shapes(shape, inShape)

        # Define shorthands.
        inLength  = inShape[0]  * inShape[1]
        outLength = outShape[0] * outShape[1]

        # If the input is stored as complex and no complex output is explicitly
        # requested, try to remove an imaginary part of zero.
        if value.typecode == "z" and typecode != "z" and not any(value.imag()):
            value = value.real()
            alwaysCopy = False

        # Broadcast or reshape the data if necessary and possible.
        reshaped = True
        if inShape == outShape:
            reshaped = False
        elif inShape == (1, 1):
            # Broadcast a scalar.
            value = load_dense(value[0], outShape, typecode)
        elif inShape == (outShape[0], 1):
            # Copy columns horizontally.
            value = load_dense(list(value)*outShape[1], outShape, typecode)
        elif inShape == (1, outShape[1]):
            # Copy rows vertically.
            value = load_dense(
                list(value)*outShape[0], (outShape[1], outShape[0]), typecode).T
        elif (inLength, 1) == outShape:
            # Vectorize in column-major order.
            if not typecode or value.typecode == typecode:
                # Quick method.
                value = value[:]
            else:
                # With typecode change.
                value = load_dense(value, outShape, typecode)
        elif (1, inLength) == outShape:
            # Vectorize in row-major order.
            if not typecode or value.typecode == typecode:
                # Quick method.
                value = value.T[:].T
            else:
                # With typecode change.
                value = load_dense(value.T, outShape, typecode)
        elif (outLength, 1) == inShape:
            # Devectorize in column-major order.
            value = load_dense(value, outShape, typecode)
        elif (1, outLength) == inShape:
            # Devectorize in row-major order.
            value = load_dense(value, (outShape[1], outShape[0]), typecode).T
        else:
            broadcast_error()
        if reshaped:
            alwaysCopy = False

        # The data now has the desired shape.
        assert value.size == outShape

        # Ensure the proper typecode.
        if typecode and value.typecode != typecode:
            value = load_dense(value, outShape, typecode)
            alwaysCopy = False

        # Copy the data if requested and not already a copy.
        if alwaysCopy:
            value = load_dense(value, outShape, typecode)
    elif isinstance(value, cvxopt.spmatrix):
        # NOTE: See case above.

        # Refine the output shape.
        inShape  = value.size
        outShape = blend_shapes(shape, inShape)

        # Define shorthands.
        inLength  = inShape[0]  * inShape[1]
        outLength = outShape[0] * outShape[1]

        # If the input is stored as complex and no complex output is explicitly
        # requested, try to remove an imaginary part of zero.
        if value.typecode == "z" and typecode != "z" and not any(value.imag()):
            value = value.real()
            alwaysCopy = False

        # Broadcast or reshape the data if necessary and possible.
        reshaped = True
        if inShape == outShape:
            reshaped = False
        elif inShape == (1, 1):
            # Broadcast a scalar.
            if value[0] != 0:
                value = load_dense(value[0], outShape, typecode)
            else:
                value = load_sparse([], [], [], outShape, typecode)
        elif inShape == (outShape[0], 1):
            # Copy columns horizontally.
            V = list(value.V)*outShape[1]
            I = list(value.I)*outShape[1]
            J = [j for j in range(outShape[1]) for _ in range(len(value))]
            value = load_sparse(V, I, J, outShape, typecode)
        elif inShape == (1, outShape[1]):
            # Copy rows vertically.
            V = list(value.V)*outShape[0]
            I = [i for i in range(outShape[0]) for _ in range(len(value))]
            J = list(value.J)*outShape[0]
            value = load_sparse(V, I, J, outShape, typecode)
        elif (inLength, 1) == outShape:
            # Vectorize in column-major order.
            if not typecode or value.typecode == typecode:
                # Quick method.
                value = value[:]
            else:
                # With typecode change.
                n, nnz = inShape[0], len(value)
                I, J = value.I, value.J
                I = [I[k] % n + J[k]*n for k in range(nnz)]
                J = [0]*nnz
                value = load_sparse(value.V, I, J, outShape, typecode)
        elif (1, inLength) == outShape:
            # Vectorize in row-major order.
            if not typecode or value.typecode == typecode:
                # Quick method.
                value = value.T[:].T
            else:
                # With typecode change.
                m, nnz = inShape[1], len(value)
                I, J = value.I, value.J
                J = [I[k]*m + J[k] % m for k in range(nnz)]
                I = [0]*nnz
                value = load_sparse(value.V, I, J, outShape, typecode)
        elif (outLength, 1) == inShape:
            # Devectorize in column-major order.
            # TODO: Logic for also changing the typecode.
            value = value[:]
            value.size = outShape
        elif (1, outLength) == inShape:
            # Devectorize in row-major order.
            # TODO: Logic for also changing the typecode.
            value = value[:]
            value.size = (outShape[1], outShape[0])
            value = value.T
        else:
            broadcast_error()
        if reshaped:
            alwaysCopy = False

        # The data now has the desired shape.
        assert value.size == outShape

        # Ensure the proper typecode.
        if typecode and value.typecode != typecode:
            # NOTE: In the case of intential loading as a dense matrix, the
            #       typecode is already set properly.
            assert isinstance(value, cvxopt.spmatrix)
            value = load_sparse(value.V, value.I, value.J, outShape, typecode)
            alwaysCopy = False

        # Return either a (sparse) copy or the input data itself.
        if isinstance(value, cvxopt.matrix):
            # The data was intentionally copied to a dense matrix (through
            # broadcasting), so only convert it back to sparse if requested.
            assert reshaped
            if sparse:
                value = cvxopt.sparse(value)
        elif sparse is False:
            value = load_dense(value, outShape, typecode)
        elif alwaysCopy:
            value = load_sparse(value.V, value.I, value.J, outShape, typecode)
    elif isinstance(value, numpy.ndarray):
        # NumPy arrays can be tensors, we don't support those.
        if len(value.shape) > 2:
            raise TypeError("PICOS does not support tensor data.")

        # Refine the output shape.
        inShape  = load_shape(value.shape)
        outShape = blend_shapes(shape, inShape)

        # Define shorthands.
        inLength  = inShape[0]  * inShape[1]
        outLength = outShape[0] * outShape[1]

        # If the input is one-dimensional, turn it into a column or row vector.
        if inShape != value.shape:
            if simple_vector_as_row(shape):
                value = value.reshape((1, inLength))
            else:
                value = value.reshape(inShape)

        # If the input is stored as complex and no complex output is explicitly
        # requested, try to remove an imaginary part of zero.
        if value.dtype.kind == "c" and typecode != "z" \
        and not numpy.iscomplex(value).any():
            value = value.real

        # Broadcast or reshape the data if necessary and possible.
        if inShape == outShape:
            pass
        elif inShape == (1, 1):
            # Broadcast a scalar.
            value = numpy.full(outShape, value)
        elif inShape == (outShape[0], 1):
            # Copy columns horizontally.
            value = value.repeat(outShape[1], 1)
        elif inShape == (1, outShape[1]):
            # Copy rows vertically.
            value = value.repeat(outShape[0], 0)
        elif (inLength, 1) == outShape:
            # Vectorize in column-major order.
            value = value.reshape(outShape, order="F")
        elif (1, inLength) == outShape:
            # Vectorize in row-major order.
            value = value.reshape(outShape, order="C")
        elif (outLength, 1) == inShape:
            # Devectorize in column-major order.
            value = value.reshape(outShape, order="F")
        elif (1, outLength) == inShape:
            # Devectorize in row-major order.
            value = value.reshape(outShape, order="C")
        else:
            broadcast_error()

        # The data now has the desired shape.
        assert value.shape == outShape

        # Decide on whether to create a dense or sparse matrix.
        outSparse = should_be_sparse(outShape, numpy.count_nonzero(value)) \
            if sparse is None else sparse

        # Convert to CVXOPT.
        if outSparse:
            I, J  = value.nonzero()
            V     = value[I, J]
            value = load_sparse(V, I, J, outShape, typecode)
        else:
            value = load_dense(value, outShape, typecode)
    elif isinstance(value, ComplexAffineExpression):
        # Must be constant.
        if not value.constant:
            raise ValueError("Cannot load the nonconstant expression {} as a "
                "constant data value.".format(value.string))

        # Retrieve a copy of the numeric value.
        value = value.value

        # NOTE: alwaysCopy=False as the value is already a copy.
        return load_data(value, shape, typecode, sparse, False, legacy)
    elif isinstance(value, str) and re.search(r"\s", value):
        value = value.strip().splitlines()
        value = [row.strip().split() for row in value]
        value = [[scalar(x, typecode) for x in row] for row in value]

        return load_data(value, shape, typecode, sparse, alwaysCopy, legacy)
    elif isinstance(value, str):
        # Verbatim matrix strings with only one element fall throught to this
        # case as they don't (have to) contain a whitespace character.
        try:
            value = scalar(value, typecode)
        except ValueError:
            pass
        else:
            return load_data(value, shape, typecode, sparse, alwaysCopy, legacy)

        regex = \
            r"^([\-0-9.+j]+)?" + "(" +\
            "e_[0-9]+(?:,[0-9]+)?" + "|" +\
            r"\|[\-0-9.+j]+\|" + "|" +\
            "I" + ")" +\
            r"(\([0-9]+(?:,[0-9]+)?\))?$"

        try:
            tokens = re.findall(regex, value)[0]
            if len(tokens) != 3:
                raise ValueError("Unexpected number of tokens.")
        except Exception:
            raise ValueError("The string '{}' could not be parsed as a matrix."
                .format(value))

        # Retrieve the string tokens.
        factor, base, inShape = tokens

        # Convert the factor.
        factor = scalar(factor) if factor else 1

        # Determine whether the matrix will be square.
        square = base == "I"

        # Convert the shape.
        if inShape:
            inShape = inShape[1:-1].split(",")
            if len(inShape) == 1:
                inShape *= 2
            inShape = (int(inShape[0]), int(inShape[1]))

            if blend_shapes(shape, inShape) != inShape:
                raise ValueError("Inconsistent shapes for matrix given as '{}' "
                    "with expected shape {}."
                    .format(value, glyphs.shape(shape)))

            outShape = inShape
        elif None in shape:
            if square and shape != (None, None):
                outShape = blend_shapes(shape, (shape[1], shape[0]))
                assert None not in outShape
            else:
                raise ValueError("Could not determine the shape of a matrix "
                    "given as '{}' because the expected size of {} contains a "
                    "wildcard. Try to give the shape explicitly with the "
                    "string.".format(value, glyphs.shape(shape)))
        else:
            outShape = shape

        # Create the base matrix.
        if base.startswith("e_"):
            position = base[2:].split(",")
            if len(position) == 1:
                index = int(position[0]) - 1
                i, j = index % outShape[0], index // outShape[0]
            else:
                i, j = int(position[0]) - 1, int(position[1]) - 1

            if i >= outShape[0] or j >= outShape[1]:
                raise ValueError("Out-of-boundary unit at row {}, column {} "
                    "in matrix of shape {} given as '{}'."
                    .format(i + 1, j + 1, glyphs.shape(outShape), value))

            value = load_sparse([1.0], [i], [j], outShape, typecode)

            if outShape[1] == 1:
                assert j == 0
                string = "e_{}".format(i + 1)
            elif outShape[0] == 1:
                assert i == 0
                string = glyphs.transp("e_{}".format(j + 1))
            else:
                string = "e_{},{}".format(i + 1, j + 1)
        elif base.startswith("|"):
            element = scalar(base[1:-1])

            # Pull the factor inside the matrix.
            element *= factor
            factor = 1

            value  = load_dense(element, outShape, typecode)
            string = glyphs.scalar(element)
        elif base == "I":
            if outShape[0] != outShape[1]:
                raise ValueError("Cannot create a non-square identy matrix.")

            n = outShape[0]

            value = load_sparse([1.0]*n, range(n), range(n), outShape, typecode)
            string = glyphs.scalar(1) if n == 1 else glyphs.idmatrix()
        else:
            assert False, "Unexpected matrix base string."

        # Apply a coefficient.
        if factor != 1:
            value *= factor
            string = glyphs.mul(glyphs.scalar(factor), string)

        # Finalize the string.
        if base.startswith("e_"):
            string = glyphs.matrix(
                glyphs.compsep(glyphs.shape(outShape), string))
        elif base.startswith("|"):
            string = glyphs.matrix(string)
        elif base == "I":
            pass
        else:
            assert False, "Unexpected matrix base string."

        # Convert between dense and sparse representation.
        if sparse is True and isinstance(value, cvxopt.matrix):
            value = cvxopt.sparse(value)
        elif sparse is False and isinstance(value, cvxopt.spmatrix):
            value = cvxopt.matrix(value)
    elif isinstance(value, (tuple, list)):
        if not value:
            raise ValueError("Cannot parse an empty tuple or list as a matrix.")

        rows   = len(value)
        cols   = None
        nested = isinstance(value[0], (tuple, list))

        if nested:
            # Both outer and inner container must be lists. Unconditionally make
            # a copy of the outer container so we can replace any inner tuples.
            value = list(value)

            for rowNum, row in enumerate(value):
                if not isinstance(row, (tuple, list)):
                    raise TypeError("Expected a tuple or list for a matrix row "
                        "but found an element of type {}.".format(
                        type(row).__name__))

                # Make sure row length is consistent.
                if cols is None:
                    cols = len(row)

                    if not cols:
                        raise TypeError("Cannot parse an empty tuple or list as"
                            " a matrix row.")
                elif len(row) != cols:
                    raise TypeError("Rows of differing size in a matrix given "
                        "as a tuple or list: {}, {}.".format(cols, len(row)))

                if isinstance(row, tuple):
                    value[rowNum] = list(row)

            value = load_dense(value, (cols, rows), typecode).T
        elif rows == 1:
            outShape = blend_shapes(shape, (1, 1))
            return load_data(
                value[0], outShape, typecode, sparse, alwaysCopy, legacy)
        else:
            outShape = (1, rows) if simple_vector_as_row(shape) else (rows, 1)
            value = load_dense(value, outShape, typecode)

        # Recurse for further transformations (broadcasting, sparsity).
        return load_data(value, shape, typecode, sparse, False, legacy)
    else:
        raise TypeError("PICOS can't load an object of type {} as a matrix: {}."
            .format(type(value).__name__, repr(value)))

    # HACK: Work around CVXOPT not supporting integer sparse matrices: If
    #       integer is requested and the matrix is currently sparse, turn dense.
    #       Note that users may not request both sparsity and integrality.
    if typecode == "i" and value.typecode == "d":
        assert not sparse
        assert isinstance(value, cvxopt.spmatrix)
        value = load_dense(value, value.size, typecode)

    if legacy:
        # Handle the case of broadcasting a scalar for matrix multiplication.
        assert inShape is not None, "Conversions must define 'inSize'."
        if inShape == (1, 1) and None in shape and shape != (None, None) \
        and 1 not in shape:
            assert 1 in value.size

            value = cvxopt.spdiag(value)

            if sparse is False:
                value = cvxopt.dense(value)

            scalarString = glyphs.scalar(value[0])
            string = glyphs.mul(scalarString, glyphs.idmatrix()) \
                if scalarString != "1" else glyphs.idmatrix()

    # Validate the output shape and type.
    assert value.size == blend_shapes(shape, value.size)
    assert not typecode or value.typecode == typecode
    if sparse is None:
        assert isinstance(value, (cvxopt.matrix, cvxopt.spmatrix))
    elif sparse:
        assert isinstance(value, cvxopt.spmatrix)
    else:
        assert isinstance(value, cvxopt.matrix)

    # Fallback to a generic matrix string if no better string was found.
    if string is None:
        if value.size == (1, 1):
            string = glyphs.scalar(value[0])
        else:
            string = glyphs.matrix(glyphs.shape(value.size))

    return value, string


def load_sparse_data(value, shape=None, typecode=None, alwaysCopy=True):
    """See :func:`~.data.load_data` with ``sparse = True``."""
    return load_data(value=value, shape=shape, typecode=typecode, sparse=True,
        alwaysCopy=alwaysCopy)


def load_dense_data(value, shape=None, typecode=None, alwaysCopy=True):
    """See :func:`~.data.load_data` with ``sparse = False``."""
    return load_data(value=value, shape=shape, typecode=typecode, sparse=False,
        alwaysCopy=alwaysCopy)


def convert_and_refine_arguments(*which, **kwargs):
    """Convert selected function arguments to PICOS expressions.

    If the selected arguments are already PICOS expressions, they are refined
    unless disabled. If they are not already PICOS expressions, an attempt is
    made to load them as constant expressions.

    :Decorator guarantee:

    All specified arguments are refined PICOS expressions when the function is
    exectued.

    :param bool refine:
        Whether to refine arguments that are already PICOS expressions.
    :param bool allowNone:
        Whether :obj:`None` is passed through to the function.
    """
    # HACK: Python 2 does not allow individual keyword arguments following
    #       starred positional ones, so map them manually.
    refine    = kwargs.pop("refine")    if "refine"    in kwargs else True
    allowNone = kwargs.pop("allowNone") if "allowNone" in kwargs else False
    if kwargs:  # Mimic Python 3's usual exception.
        raise TypeError("'{}' is an invalid keyword argument for "
            "convert_and_refine_arguments()".format(next(kwargs.keys())))

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            from .expression import Expression
            from .exp_affine import Constant

            def name():
                if hasattr(func, "__qualname__"):
                    return func.__qualname__
                else:
                    return func.__name__

            callargs = inspect.getcallargs(func, *args, **kwargs)

            newargs = {}
            for key, arg in callargs.items():
                if key not in which:
                    pass
                elif allowNone and arg is None:
                    pass
                elif isinstance(arg, Expression):
                    if refine:
                        arg = arg.refined
                else:
                    try:
                        arg = Constant(arg)
                    except Exception as e:
                        raise TypeError("Failed to convert argument '{}' of {} "
                            "to a PICOS constant: {}".format(key, name(), e))

                newargs[key] = arg

            return func(**newargs)
        return wrapper
    return decorator


def convert_operands(
        sameShape=False, scalarLHS=False, scalarRHS=False, rMatMul=False,
        lMatMul=False, horiCat=False, vertCat=False, allowNone=False):
    """Convert binary operator operands to PICOS expressions.

    A decorator for a binary operator that converts any operand that is not
    already a PICOS expression or set into a constant one that fulfills the
    given shape requirements, if possible. See :func:`~.data.load_data` for
    broadcasting and reshaping rules that apply to raw data.

    If both operands are already PICOS expressions and at least one of them is
    an affine expression, there is a limited set of broadcasting rules to fix
    a detected shape mismatch. If this does not succeed, an exception is raised.
    If no operand is affine, the operation is performed even if shapes do not
    match. The operation is responsible for dealing with this case.

    If either operand is a PICOS set, no broadcasting or reshaping is applied as
    set instances have, in general, variable dimensionality. If a set type can
    *not* have arbitrary dimensions, then it must validate the element's shape
    on its own. In particular, no shape requirement may be given when this
    decorator is used on a set method.

    :Decorator guarantee:

    This decorator guarantees to the binary operator using it that only PICOS
    expression or set types will be passed as operands and that any affine
    expression already has the proper shape for the operation, based on the
    decorator arguments.

    :Broadcasting rules for affine expressions:

    Currently, only scalar affine expressions are broadcasted to the next
    smallest matching shape. This is more limited than the broadcasting behavior
    when one of the operands is raw data, but it ensures a symmetric behavior in
    case both operands are affine.

    :param bool sameShape: Both operands must have the exact same shape.
    :param bool scalarLHS: The left hand side operand must be scalar.
    :param bool scalarRHS: The right hand side operand must be scalar.
    :param bool rMatMul: The operation has the shape requirements of normal
        matrix multiplication with the second operand on the right side.
    :param bool lMatMul: The operation has the shape requirements of reversed
        matrix multiplication with the second operand on the left side.
    :param bool horiCat: The operation has the shape requirements of horizontal
        matrix concatenation.
    :param bool vertCat: The operation has the shape requirements of vertical
        matrix concatenation.
    :param bool allowNone: An operand of :obj:`None` is passed as-is.

    :raises TypeError: If matching shapes cannot be produced despite one of the
        operands being raw data or an affine expression.

    .. note::
        Matrix multiplication includes scalar by matrix multiplication, so
        either operand may remain a scalar.
    """
    # Fix a redundancy in allowed argument combinations.
    if sameShape and (scalarLHS or scalarRHS):
        sameShape = False
        scalarLHS = True
        scalarRHS = True

    # Check arguments, only scalarLHS and scalarRHS may appear together.
    anyScalar = scalarLHS or scalarRHS
    selected  = len([arg for arg in
        (sameShape, anyScalar, lMatMul, rMatMul, horiCat, vertCat) if arg])
    if selected > 1:
        assert False, "Conflicting convert_operands arguments."

    def decorator(operator):
        @functools.wraps(operator)
        def wrapper(lhs, rhs, *args, **kwargs):
            def error(reason="Operand shapes do not match."):
                opName  = operator.__qualname__ \
                    if hasattr(operator, "__qualname__") else operator.__name__
                lhsName = lhs.string if hasattr(lhs, "string") else repr(lhs)
                rhsName = rhs.string if hasattr(rhs, "string") else repr(rhs)
                raise TypeError("Invalid operation {}({}, {}): {}".format(
                    opName, lhsName, rhsName, reason))

            def make_lhs_shape(rhsShape):
                if sameShape:
                    return rhsShape
                elif horiCat:
                    return (rhsShape[0], None)
                elif vertCat:
                    return (None, rhsShape[1])
                elif rMatMul:
                    return (None, rhsShape[0])
                elif lMatMul:
                    return (rhsShape[1], None)
                elif scalarLHS:
                    return (1, 1)
                else:
                    return None

            def make_rhs_shape(lhsShape):
                if sameShape:
                    return lhsShape
                elif horiCat:
                    return (lhsShape[0], None)
                elif vertCat:
                    return (None, lhsShape[1])
                elif rMatMul:
                    return (lhsShape[1], None)
                elif lMatMul:
                    return (None, lhsShape[0])
                elif scalarRHS:
                    return (1, 1)
                else:
                    return None

            from .set        import Set
            from .expression import Expression
            from .exp_affine import Constant, ComplexAffineExpression

            lhsIsExpOrSet = isinstance(lhs, (Expression, Set))
            rhsIsExpOrSet = isinstance(rhs, (Expression, Set))

            lhsIsSet = lhsIsExpOrSet and isinstance(lhs, Set)
            rhsIsSet = rhsIsExpOrSet and isinstance(rhs, Set)

            if lhsIsSet:
                assert not selected, "convert_operands when used on sets may " \
                    "not pose any shape requirements."

            if lhsIsExpOrSet and rhsIsExpOrSet:
                lhsIsAffine = isinstance(lhs, ComplexAffineExpression)
                rhsIsAffine = isinstance(rhs, ComplexAffineExpression)

                # If neither expression is affine, it's the operation's job to
                # deal with it.
                if not lhsIsAffine and not rhsIsAffine:
                    return operator(lhs, rhs, *args, **kwargs)

                # If there are no shape requirements, we are done as both are
                # already expressions.
                if not selected:
                    return operator(lhs, rhs, *args, **kwargs)

                assert not lhsIsSet  # Handled earlier.

                # Sets have variable shape, so no adjustment is necessary.
                if rhsIsSet:
                    return operator(lhs, rhs, *args, **kwargs)

                lhsShape, rhsShape = lhs.shape, rhs.shape

                # Check if already matching size.
                if (sameShape or horiCat or vertCat) and lhsShape == rhsShape:
                    return operator(lhs, rhs, *args, **kwargs)

                lm, ln = lhs.shape
                rm, rn = rhs.shape

                # Further check if already matching size.
                if (horiCat and lm == rm) \
                or (vertCat and ln == rn) \
                or (rMatMul and ln == rm) \
                or (lMatMul and rn == lm):
                    return operator(lhs, rhs, *args, **kwargs)

                lhsL, rhsL = len(lhs), len(rhs)

                # scalarLHS and scalarRHS are the only two shape requirements
                # that may appear together, so handle all combinations here.
                if scalarLHS and scalarRHS:
                    if lhsL == 1 and rhsL == 1:
                        return operator(lhs, rhs, *args, **kwargs)
                elif scalarLHS and lhsL == 1:
                    return operator(lhs, rhs, *args, **kwargs)
                elif scalarRHS and rhsL == 1:
                    return operator(lhs, rhs, *args, **kwargs)

                # Matrix multiplication always accepts scalars.
                if (rMatMul or lMatMul) and 1 in (lhsL, rhsL):
                    return operator(lhs, rhs, *args, **kwargs)

                # Broadcast an affine scalar on the left side to match.
                if lhsIsAffine and lhsL == 1:
                    lhs = lhs.broadcasted(make_lhs_shape(rhsShape))
                    return operator(lhs, rhs, *args, **kwargs)

                # Broadcast an affine scalar on the right side to match.
                if rhsIsAffine and rhsL == 1:
                    rhs = rhs.broadcasted(make_rhs_shape(lhsShape))
                    return operator(lhs, rhs, *args, **kwargs)

                # At least one of the expressions is affine and we didn't find a
                # way to fix a detected shape mismatch. It's our job to error.
                error("The operand shapes of {} and {} do not match."
                    .format(glyphs.shape(lhsShape), glyphs.shape(rhsShape)))
            elif lhsIsExpOrSet:
                if allowNone and rhs is None:
                    return operator(lhs, rhs, *args, **kwargs)

                rhsShape = None if lhsIsSet else make_rhs_shape(lhs.shape)

                try:
                    if rMatMul or lMatMul:
                        if lhs.shape == (1, 1):
                            # Any shape works.
                            rhs = Constant(rhs)
                        else:
                            try:
                                rhs = Constant(rhs, shape=(1, 1))
                            except Exception:
                                rhs = Constant(rhs, shape=rhsShape)
                    else:
                        rhs = Constant(rhs, shape=rhsShape)
                except Exception as exc:
                    error("Failed to load right hand side as a constant of "
                        "shape {}: {}".format(glyphs.shape(rhsShape), exc))
            elif rhsIsExpOrSet:
                if allowNone and lhs is None:
                    return operator(lhs, rhs, *args, **kwargs)

                lhsShape = None if rhsIsSet else make_lhs_shape(rhs.shape)

                try:
                    if rMatMul or lMatMul:
                        if rhs.shape == (1, 1):
                            # Any shape works.
                            lhs = Constant(lhs)
                        else:
                            try:
                                lhs = Constant(lhs, shape=(1, 1))
                            except TypeError:
                                lhs = Constant(lhs, shape=lhsShape)
                    else:
                        lhs = Constant(lhs, shape=lhsShape)
                except Exception as exc:
                    error("Failed to load left hand side as a constant of "
                        "shape {}: {}".format(glyphs.shape(rhsShape), exc))
            else:
                assert False, "convert_operands is supposed to decorate " \
                    "expression methods, but neither operand is a PICOS " \
                    "expression or set."

            return operator(lhs, rhs, *args, **kwargs)
        return wrapper
    return decorator


def cvxopt_equals(A, B, tolerance=None):
    """Whether two CVXOPT (sparse) matrices are numerically close.

    :param float tolerance: Maximum entrywise absolute deviation allowed.
        Defaults to :data:`TOLERANCE`.
    """
    if tolerance is None:
        tolerance = TOLERANCE

    if A.size != B.size:
        return False

    # Work around "ValueError: max() arg is an empty sequence" for sparse zero.
    if not A and not B:
        return True

    return max(abs(A - B)) <= tolerance


def cvxopt_maxdiff(A, B):
    """Return the largest absolute difference of two (sparse) CVXOPT matrices.

    :raises TypeError: If the matrices are not of the same shape.
    """
    if A.size != B.size:
        raise TypeError("The matrices do not have the same shape.")

    # Work around "ValueError: max() arg is an empty sequence" for sparse zero.
    if not A and not B:
        return 0.0

    return max(abs(A - B))


def cvxopt_hcat(matrices):
    """Concatenate the given CVXOPT (sparse) matrices horizontally.

    The resulting matrix is sparse if any input matrix is sparse.

    :param list matrices: A list of CVXOPT (sparse) matrices.
    """
    if not isinstance(matrices, list):
        matrices = list(matrices)

    sparse = any(isinstance(M, cvxopt.spmatrix) for M in matrices)

    matrices = [[matrix] for matrix in matrices]

    if sparse:
        return cvxopt.sparse(matrices)
    else:
        return cvxopt.matrix(matrices)


def cvxopt_vcat(matrices):
    """Concatenate the given CVXOPT (sparse) matrices vertically.

    The resulting matrix is sparse if any input matrix is sparse.

    :param list matrices: A list of CVXOPT (sparse) matrices.
    """
    if not isinstance(matrices, list):
        matrices = list(matrices)

    if any(isinstance(M, cvxopt.spmatrix) for M in matrices):
        return cvxopt.sparse(matrices)
    else:
        return cvxopt.matrix(matrices)


def cvxopt_hpsd(matrix):
    """Whether the given CVXOPT matrix is hermitian positive semidefinite.

    Note that the matrix must be exactly hermitian, no tolerance is applied.
    """
    if matrix.size[0] != matrix.size[1]:
        return False

    if not cvxopt_equals(matrix, matrix.H, tolerance=0.0):
        return False

    eigenvalues = numpy.linalg.eigvalsh(cvx2np(matrix))

    return all(ev >= 0 for ev in eigenvalues)


def sparse_quadruple(A, reshape=None, preT=False, postT=False):
    """Return a sparse representation of the given CVXOPT (sparse) matrix.

    :param reshape: If set, then :math:`A` is reshaped on the fly.
    :param bool preT: Transpose :math:`A` before reshaping.
    :param bool postT: Transpose :math:`A` after reshaping.

    :returns: A quadruple of values, row indices, column indices, and shape.
    """
    if not isinstance(A, (cvxopt.spmatrix, cvxopt.matrix)):
        raise TypeError("Input must be a CVXOPT (sparse) matrix.")

    m, n = A.size

    if reshape:
        reshape = load_shape(reshape)

        if A.size[0] * A.size[1] != reshape[0] * reshape[1]:
            raise TypeError("Cannot reshape from {} to {}.".format(
                glyphs.shape(A.size), glyphs.shape(reshape)))

    if not A:
        V, I, J = [], [], []
        p, q = reshape if reshape else (n, m if preT else m, n)
    elif isinstance(A, cvxopt.spmatrix):
        V, I, J = A.V, A.I, A.J

        if preT:
            I, J, m, n = J, I, n, m

        if reshape:
            p, q = reshape
            I, J = zip(*[((i + j*m) % p, (i + j*m) // p) for i, j in zip(I, J)])
        else:
            p, q = m, n

        V, I, J = tuple(V), tuple(I), tuple(J)
    else:
        if reshape:
            p, q = reshape
            if preT:
                V, I, J = zip(*[
                    (v, (k // m + (k % m)*n) % p, (k // m + (k % m)*n) // p)
                    for k, v in enumerate(A) if v])
            else:
                V, I, J = zip(
                    *[(v, k % p, k // p) for k, v in enumerate(A) if v])
        else:
            if preT:
                p, q = n, m
                V, I, J = zip(
                    *[(v, k // m, k % m) for k, v in enumerate(A) if v])
            else:
                p, q = m, n
                V, I, J = zip(
                    *[(v, k % m, k // m) for k, v in enumerate(A) if v])

    if postT:
        I, J, p, q = J, I, q, p

    return V, I, J, (p, q)


def left_kronecker_I(A, k, reshape=None, preT=False, postT=False):
    r"""Return :math:`I_k \otimes A` for a CVXOPT (sparse) matrix :math:`A`.

    In other words, if :math:`A` is a :math:`m \times n` CVXOPT (sparse)
    matrix, returns a :math:`km \times kn` CVXOPT sparse block matrix with
    all blocks of size :math:`m \times n`, the diagonal blocks (horizontal
    block index equal to vertical block index) equal to :math:`A`, and all
    other blocks zero.

    :param reshape: If set, then :math:`A` is reshaped on the fly.
    :param bool preT: Transpose :math:`A` before reshaping.
    :param bool postT: Transpose :math:`A` after reshaping.

    :returns:
        If :math:`A` is dense and :math:`k = 1`, a
        :func:`CVXOPT dense matrix <cvxopt:cvxopt.matrix>`, otherwise a
        :func:`CVXOPT sparse matrix <cvxopt:cvxopt.spmatrix>`.
    """
    A = A.T if preT else A

    if reshape:
        A = A if preT else A[:]  # Copy if not already fresh.
        A.size = reshape

    A = A.T if postT else A

    if k == 1:
        if not preT and not reshape and not postT:
            return A + 0  # Always return a copy.
        else:
            return A

    if A.size[0] == A.size[1]:
        A = cvxopt.spdiag([A for _ in range(k)])
    else:
        Z = cvxopt.spmatrix([], [], [], A.size)
        A = cvxopt.sparse([[Z]*i + [A] + [Z]*(k-i-1) for i in range(k)])

    return A.T if postT else A


def right_kronecker_I(A, k, reshape=None, preT=False, postT=False):
    r"""Return :math:`A \otimes I_k` for a CVXOPT (sparse) matrix :math:`A`.

    :param reshape: If set, then :math:`A` is reshaped on the fly.
    :param bool preT: Transpose :math:`A` before reshaping.
    :param bool postT: Transpose :math:`A` after reshaping.

    :returns:
        If :math:`A` is dense and :math:`k = 1`, a
        :func:`CVXOPT dense matrix <cvxopt:cvxopt.matrix>`, otherwise a
        :func:`CVXOPT sparse matrix <cvxopt:cvxopt.spmatrix>`.
    """
    if isinstance(A, cvxopt.matrix):
        # Dense case: Use NumPy.
        A = numpy.array(A)

        A = A.T if preT else A
        A = A.reshape(reshape, order="F") if reshape else A
        A = A.T if postT else A

        A = numpy.kron(A, numpy.eye(k))

        return load_data(A, sparse=(k > 1))[0]
    else:
        # Sparse case: Python implementation.
        # This is slower than the NumPy approach in general but can handle
        # arbitrarily large matrices given that they are sufficiently sparse.
        V, I, J, shape = sparse_quadruple(A, reshape, preT, postT)
        m, n = shape

        if V:
            V, I, J = zip(*[(v, k*i + l, k*j + l)
                for v, i, j in zip(V, I, J) for l in range(k)])

        return cvxopt.spmatrix(V, I, J, (k*m, k*n))


def cvx2np(A, reshape=None):
    """Convert a CVXOPT (sparse) matrix to a NumPy matrix.

    :param A: The CVXOPT :func:`dense <cvxopt:cvxopt.matrix>` or
        :func:`sparse <cvxopt:cvxopt.spmatrix>` matrix to convert.
    :param bool reshape: Optional new shape for the converted matrix.

    :returns: Converted :class:`NumPy matrix <numpy.matrix>`.
    """
    assert isinstance(A, (cvxopt.matrix, cvxopt.spmatrix))

    if isinstance(A, cvxopt.spmatrix):
        A = cvxopt.matrix(A)

    if reshape:
        shape = load_shape(reshape)
        return numpy.reshape(A, shape, "F")
    else:
        return numpy.matrix(A)


def make_fraction(p, denominator_limit):
    """Convert a float :math:`p` to a limited precision fraction.

    :param float p: The float to convert, may be positive or negative infinity.
    :param int denominator_limit: The largest allowed denominator.

    :returns tuple: A quadruple ``(num, den, pNew, pStr)`` with ``pNew`` the
        limited precision version of :math:`p`, ``pStr`` a string representation
        of the fraction, and ``num`` and ``den`` the numerator and the
        denominator of the fraction, respectively.
    """
    # LEGACY: Old tools.tracepow allowed this.
    if p in ("inf", "-inf"):
        p = float(p)

    if p in (float("inf"), float("-inf")):
        return p, 1, p, glyphs.neg(glyphs.infty) if p < 0 else glyphs.infty

    frac = Fraction(p).limit_denominator(denominator_limit)
    num  = frac.numerator
    den  = frac.denominator
    pNew = float(num) / float(den)

    if den == 1:
        pStr = glyphs.scalar(num)
    else:
        pStr = glyphs.clever_div(glyphs.scalar(num), glyphs.scalar(den))

    return num, den, pNew, pStr


def value(obj, sparse=None, numpy=False):
    """Convert (nested) PICOS objects to their current value.

    :param obj:
        Either a single (PICOS) object that has a ``value`` attribute, such as a
        :class:`variable <picos.expressions.variables.BaseVariable>`,
        :class:`expression <picos.expressions.expression.Expression>` or
        :class:`~picos.modeling.Problem`, or a (nested) :class:`list`,
        :class:`tuple` or :class:`dict` thereof.

    :param sparse:
        If :obj:`None`, retrieved multidimensional values can be returned as
        either CVXOPT :func:`sparse <cvxopt:cvxopt.spmatrix>` or
        :func:`dense <cvxopt:cvxopt.matrix>` matrices, whichever PICOS stores
        internally. If :obj:`True` or :obj:`False`, multidimensional values are
        always returned as sparse or dense types, respectively.

    :param bool numpy:
        If :obj:`True`, retrieved multidimensional values are returned as a
        NumPy :class:`~numpy:numpy.ndarray` instead of a CVXOPT type. May not be
        set in combination with ``sparse=True``.

    :returns:
        An object of the same (nested) structure as ``obj``, with every
        occurence of any object with a ``value`` attribute replaced by that
        attribute's current numeric value. In the case of dictionaries, only the
        dictionary values will be converted.

    :raises TypeError:
        If some object with a ``value`` attribute has a value that cannot be
        converted to a matrix by :func:`~.load_data`. This can only happen if
        the object in question is not a PICOS object.

    :Example:

    >>> from picos import RealVariable, value
    >>> from pprint import pprint
    >>> x = {key: RealVariable(key) for key in ("foo", "bar")}
    >>> x["foo"].value = 2
    >>> x["bar"].value = 3
    >>> pprint(value(x))
    {'bar': 3.0, 'foo': 2.0}
    """
    if sparse and numpy:
        raise ValueError("NumPy does not support sparse matrices.")

    if isinstance(obj, tuple):
        return tuple(value(inner, sparse, numpy) for inner in obj)
    elif isinstance(obj, list):
        return [value(inner, sparse, numpy) for inner in obj]
    elif isinstance(obj, dict):
        return {k: value(v, sparse, numpy) for k, v in obj.items()}
    else:
        if hasattr(obj, "value"):
            val = obj.value

            if isinstance(val, (int, float, complex)):
                return val

            # PICOS objects always return their value as a CVXOPT matrix type,
            # but this function may be used on other objects with a value
            # attribute. Try to convert their value to a CVXOPT matrix first.
            if not isinstance(val, (cvxopt.matrix, cvxopt.spmatrix)):
                if numpy:
                    load_sparse = False
                elif sparse:
                    load_sparse = True
                else:
                    load_sparse = None

                val = load_data(val, sparse=load_sparse)[0]
            elif isinstance(val, cvxopt.spmatrix) \
            and (sparse is False or numpy):
                val = cvxopt.dense(val)

            if numpy:
                import numpy as the_numpy
                assert isinstance(val, cvxopt.matrix)
                val = the_numpy.array(val)

            return val
        else:
            return obj


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`ECOSSolver`."""

import cvxopt

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import (AffineConstraint, ExpConeConstraint, RSOCConstraint,
                           SOCConstraint)
from ..expressions import AffineExpression, BinaryVariable, IntegerVariable
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_INFEASIBLE, PS_UNBOUNDED,
                                 PS_UNKNOWN, PS_UNSTABLE, SS_FAILURE,
                                 SS_INFEASIBLE, SS_OPTIMAL, SS_PREMATURE,
                                 SS_UNKNOWN)
from .solver import Solver

_API_START = api_start(globals())
# -------------------------------


class ECOSSolver(Solver):
    """Interface to the ECOS solver via its official Python interface."""

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression],
        variables=Specification.ALL,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            ExpConeConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 1.0  # Stable free/open source solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("ecos")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "ecos", "ECOS", "Embedded Conic Solver"

    def __init__(self, problem):
        """Initialize an ECOS solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        super(ECOSSolver, self).__init__(problem)

        self._numVars = 0
        """Total number of scalar variables passed to ECOS."""

        self._ecosVarOffset = {}
        """Maps a PICOS variable to its offset in the ECOS constraint matrix."""

        self._ecosConIndices = dict()
        """Maps a PICOS constraints to a range of ECOS constraint indices.

        Note that equality constraints use a different index space."""

        self._ecosModuleCache = None
        """A cached reference to the ECOS module."""

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        self.int = None

        self._numVars = 0
        self._ecosVarOffset.clear()
        self._ecosConIndices.clear()

        # self._objectiveOffset = 0.0

    @property
    def ecos(self):
        """Return the ECOS core module ecos.py.

        The module is obtained by ``import ecos`` up to ECOS 2.0.6 and by
        ``import ecos.ecos`` starting with ECOS 2.0.7.
        """
        if self._ecosModuleCache is None:
            import ecos
            if hasattr(ecos, "ecos"):
                self._ecosModuleCache = ecos.ecos
            else:
                self._ecosModuleCache = ecos
        return self._ecosModuleCache

    @property
    def array(self):
        """ECOS' array type."""
        return self.ecos.np.array

    @property
    def matrix(self):
        """ECOS' matrix type."""
        return self.ecos.sparse.csc_matrix

    def zeros(self, shape):
        """Create a zero array or a zero matrix, depending on ``shape``."""
        if isinstance(shape, int) or len(shape) == 1:
            return self.ecos.np.zeros(shape)
        else:
            return self.matrix(shape)

    def stack(self, *args):
        """Stack vectors or matrices, the latter vertically."""
        # In the case of matrices, stack vertically.
        if isinstance(args[0], self.ecos.sparse.base.spmatrix):
            for i in range(1, len(args)):
                assert isinstance(args[i], self.ecos.sparse.base.spmatrix)
            return self.ecos.sparse.vstack(args, format="csc")

        # In the case of arrays, append them.
        for i in range(len(args)):
            assert isinstance(args[i], self.ecos.np.ndarray)
        return self.ecos.np.hstack(args)

    def _affine_expression_to_G_and_h(self, expression):
        assert isinstance(expression, AffineExpression)

        length = len(expression)

        # Construct G.
        I, J, V = [], [], []
        for variable, coefficients in expression._coefs.items():
            ecosVarOffset = self._ecosVarOffset[variable]

            if not isinstance(coefficients, cvxopt.spmatrix):
                coefficients = cvxopt.sparse(coefficients)

            I.extend(coefficients.I)
            J.extend([ecosVarOffset + j for j in coefficients.J])
            V.extend(coefficients.V)
        G = self.matrix((V, (I, J)), (length, self._numVars))

        # Construct h.
        constant = expression._const
        if not isinstance(constant, cvxopt.matrix):
            constant = cvxopt.matrix(constant)
        h = self.array(constant, dtype=float).flatten()

        return G, h

    _Gh = _affine_expression_to_G_and_h

    def _import_variables(self):
        offset = 0
        for variable in self.ext.variables.values():
            dim = variable.dim

            # Mark integer variables.
            if isinstance(variable, (BinaryVariable, IntegerVariable)):
                ecosIndices = range(offset, offset + dim)

                if isinstance(variable, BinaryVariable):
                    self.int["bool_vars_idx"].extend(ecosIndices)
                else:
                    self.int["int_vars_idx"].extend(ecosIndices)

            # Register the variable.
            self._ecosVarOffset[variable] = offset
            offset += dim

            # Import bounds.
            bounds = variable.bound_constraint
            if bounds:
                self._import_affine_constraint(bounds)

        assert offset == self._numVars

    def _append_equality(self, A, b):
        offset = len(self.int["b"])
        length = len(b)

        self.int["A"] = self.stack(self.int["A"], A)
        self.int["b"] = self.stack(self.int["b"], b)

        return range(offset, offset + length)

    def _append_inequality(self, G, h, typecode):
        assert typecode in self.int["dims"]

        # Make sure constraints are appened in the proper order.
        if typecode == "l":
            assert len(self.int["dims"]["q"]) == 0 \
                and self.int["dims"]["e"] == 0
        elif typecode == "q":
            assert self.int["dims"]["e"] == 0

        offset = len(self.int["h"])
        length = len(h)

        self.int["G"] = self.stack(self.int["G"], G)
        self.int["h"] = self.stack(self.int["h"], h)

        if typecode == "q":
            self.int["dims"][typecode].append(length)
        elif typecode == "e":
            self.int["dims"][typecode] += 1
        else:
            self.int["dims"][typecode] += length

        return range(offset, offset + length)

    def _import_affine_constraint(self, constraint):
        assert isinstance(constraint, AffineConstraint)

        (G_smaller, h_smaller) = self._Gh(constraint.smaller)
        (G_greater, h_greater) = self._Gh(constraint.greater)
        G = G_smaller - G_greater
        h = h_greater - h_smaller

        if constraint.is_equality():
            return self._append_equality(G, h)
        else:
            return self._append_inequality(G, h, "l")

    def _import_socone_constraint(self, constraint):
        assert isinstance(constraint, SOCConstraint)

        (A, b) = self._Gh(constraint.ne)
        (c, d) = self._Gh(constraint.ub)

        return self._append_inequality(
            self.stack(-c, -A), self.stack(d, b), "q")

    def _import_rscone_constraint(self, constraint):
        assert isinstance(constraint, RSOCConstraint)

        (A,  b)  = self._Gh(constraint.ne)
        (c1, d1) = self._Gh(constraint.ub1)
        (c2, d2) = self._Gh(constraint.ub2)

        return self._append_inequality(
            self.stack(-(c1 + c2), -2.0*A, c2 - c1),
            self.stack(d1 + d2,     2.0*b, d1 - d2), "q")

    def _import_expcone_constraint(self, constraint):
        assert isinstance(constraint, ExpConeConstraint)

        (Gx, hx) = self._Gh(constraint.x)
        (Gy, hy) = self._Gh(constraint.y)
        (Gz, hz) = self._Gh(constraint.z)

        # ECOS' exponential cone is cl{(x,y,z) | exp(x/z) <= y/z, z > 0},
        # PICOS' is cl{(x,y,z) | x >= y*exp(z/y), y > 0}. Note that given y > 0
        # it is x >= y*exp(z/y) if and only if exp(z/y) <= x/y. Therefor we can
        # transform from our coordinates to theirs with the mapping
        # (x, y, z) ↦ (z, x, y). Further, G and h with G = (Gx, Gy, Gz) and
        # h = (hx, hy, hz) are such that G*X + h = (x, y, z) where X is the
        # row-vectorization of all variables. ECOS however expects G and h such
        # that h - G*X is constrained to be in the exponential cone.
        return self._append_inequality(
            -self.stack(Gz, Gx, Gy), self.stack(hz, hx, hy), "e")

    def _import_objective(self):
        direction, objective = self.ext.objective

        # ECOS does not support a constant offset.
        # self._objectiveOffset = objective._const[0]

        if direction != "find":
            # ECOS only supports minimization; flip the sign for maximization.
            sign = 1.0 if direction == "min" else -1.0

            # Import coefficients.
            for variable, coefficient in objective._coefs.items():
                for localIndex in range(variable.dim):
                    index = self._ecosVarOffset[variable] + localIndex
                    self.int["c"][index] = sign * coefficient[localIndex]

    def _import_problem(self):
        self._numVars = sum(var.dim for var in self.ext.variables.values())

        # ECOS' internal problem representation is stateless; a number of
        # vectors and matrices is supplied each time a search is started.
        # These vectors and matrices are thus stored in self.int.
        self.int = {
            # Objective function coefficients.
            "c": self.zeros(self._numVars),

            # Linear equality left hand side.
            "A": self.matrix((0, self._numVars)),

            # Linear equality right hand side.
            "b": self.array([]),

            # Conic inequality left hand side.
            "G": self.matrix((0, self._numVars)),

            # Conic inequality right hand side.
            "h": self.array([]),

            # Cone definition: Linear, second order, exponential dimensions.
            "dims": {"l": 0, "q": [], "e": 0},

            # Boolean variable indices.
            "bool_vars_idx": [],

            # Integer variable indices.
            "int_vars_idx": []
        }

        # NOTE: Constraints need to be append ordered by type.
        # TODO: Add fast constraints-by-type iterators to Problem.

        # Import variables with their bounds as affine constraints.
        self._import_variables()

        # Import affine constraints.
        for constraint in self.ext.constraints.values():
            if isinstance(constraint, AffineConstraint):
                self._ecosConIndices[constraint] = \
                    self._import_affine_constraint(constraint)

        # Import second order cone constraints.
        for constraint in self.ext.constraints.values():
            if isinstance(constraint, SOCConstraint):
                self._ecosConIndices[constraint] = \
                    self._import_socone_constraint(constraint)

        # Import rotated second order cone constraints.
        for constraint in self.ext.constraints.values():
            if isinstance(constraint, RSOCConstraint):
                self._ecosConIndices[constraint] = \
                    self._import_rscone_constraint(constraint)

        # Import exponential cone constraints.
        for constraint in self.ext.constraints.values():
            if isinstance(constraint, ExpConeConstraint):
                self._ecosConIndices[constraint] = \
                    self._import_expcone_constraint(constraint)

        # Make sure that no unsupported constraints are present.
        for constraint in self.ext.constraints.values():
            assert isinstance(constraint, (AffineConstraint, SOCConstraint,
                RSOCConstraint, ExpConeConstraint))

        # Set objective.
        self._import_objective()

    def _update_problem(self):
        raise NotImplementedError

    def _solve(self):
        ecosOptions = {}

        # verbosity
        beVerbose = self.verbosity() >= 1
        ecosOptions["verbose"]    = beVerbose
        ecosOptions["mi_verbose"] = beVerbose

        # rel_prim_fsb_tol, rel_dual_fsb_tol
        feasibilityTols = [tol for tol in (self.ext.options.rel_prim_fsb_tol,
                self.ext.options.rel_dual_fsb_tol) if tol is not None]
        if feasibilityTols:
            ecosOptions["feastol"] = min(feasibilityTols)

        # abs_ipm_opt_tol
        if self.ext.options.abs_ipm_opt_tol is not None:
            ecosOptions["abstol"] = self.ext.options.abs_ipm_opt_tol

        # rel_ipm_opt_tol
        if self.ext.options.rel_ipm_opt_tol is not None:
            ecosOptions["reltol"] = self.ext.options.rel_ipm_opt_tol

        # abs_bnb_opt_tol
        if self.ext.options.abs_bnb_opt_tol is not None:
            ecosOptions["mi_abs_eps"] = self.ext.options.abs_bnb_opt_tol

        # rel_bnb_opt_tol
        if self.ext.options.rel_bnb_opt_tol is not None:
            ecosOptions["mi_rel_eps"] = self.ext.options.rel_bnb_opt_tol

        # integrality_tol
        # HACK: ECOS_BB uses ECOS' "tolerance for feasibility condition for
        #       inaccurate solution" as the integrality tolerance.
        if self.ext.options.integrality_tol is not None:
            ecosOptions["feastol_inacc"] = self.ext.options.integrality_tol

        # max_iterations
        if self.ext.options.max_iterations is not None:
            ecosOptions["max_iters"]    = self.ext.options.max_iterations
            ecosOptions["mi_max_iters"] = self.ext.options.max_iterations

        # Handle unsupported options.
        self._handle_unsupported_options(
            "lp_root_method", "lp_node_method", "timelimit", "treememory",
            "max_fsb_nodes", "hotstart")

        # Assemble the solver input.
        arguments = {}
        arguments.update(self.int)
        arguments.update(ecosOptions)

        # Debug print the solver input.
        if self._debug():
            from pprint import pformat
            self._debug("Passing to ECOS:\n" + pformat(arguments))

        # Attempt to solve the problem.
        with self._header(), self._stopwatch():
            solution = self.ecos.solve(**arguments)

        # Debug print the solver output.
        if self._debug():
            from pprint import pformat
            self._debug("Recevied from ECOS:\n" + pformat(solution))

        # Retrieve primals.
        primals = {}
        if self.ext.options.primals is not False:
            for variable in self.ext.variables.values():
                offset = self._ecosVarOffset[variable]
                dim    = variable.dim
                value  = list(solution["x"][offset:offset + dim])
                primals[variable] = value

        # Retrieve duals.
        duals = {}
        if self.ext.options.duals is not False:
            for constraint in self.ext.constraints.values():
                dual    = None
                indices = self._ecosConIndices[constraint]

                if isinstance(constraint, AffineConstraint):
                    if constraint.is_equality():
                        dual = list(solution["y"][indices])
                    else:
                        dual = list(solution["z"][indices])
                elif isinstance(constraint, SOCConstraint):
                    dual     = solution["z"][indices]
                    dual     = list(dual)
                elif isinstance(constraint, RSOCConstraint):
                    dual     = solution["z"][indices]
                    dual[1:] = -dual[1:]
                    alpha    = dual[0] - dual[-1]
                    beta     = dual[0] + dual[-1]
                    z        = list(-2.0 * dual[1:-1])
                    dual     = [alpha, beta] + z
                elif isinstance(constraint, ExpConeConstraint):
                    zxy  = solution["z"][indices]
                    dual = [zxy[1], zxy[2], zxy[0]]
                else:
                    assert False, "Unexpected constraint type."

                if type(dual) is list:
                    if len(dual) == 1:
                        dual = float(dual[0])
                    else:
                        dual = cvxopt.matrix(dual, constraint.size)

                duals[constraint] = dual

        # # Retrieve objective value.
        # p = solution["info"]["pcost"]
        # d = solution["info"]["dcost"]
        #
        # if p is not None and d is not None:
        #     objectiveValue = 0.5 * (p + d)
        # elif p is not None:
        #     objectiveValue = p
        # elif d is not None:
        #     objectiveValue = d
        # else:
        #     objectiveValue = None
        #
        # if objectiveValue is not None:
        #     objectiveValue += self._objectiveOffset
        #
        #     if self.ext.objective.direction == "max":
        #         objectiveValue = -objectiveValue

        statusCode = solution["info"]["exitFlag"]
        if   statusCode == 0:  # optimal
            primalStatus   = SS_OPTIMAL
            dualStatus     = SS_OPTIMAL
            problemStatus  = PS_FEASIBLE
        elif statusCode == 10:  # inaccurate/optimal
            primalStatus   = SS_OPTIMAL
            dualStatus     = SS_OPTIMAL
            problemStatus  = PS_FEASIBLE
        elif statusCode == 1:  # primal infeasible
            primalStatus   = SS_INFEASIBLE
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_INFEASIBLE
        elif statusCode == 11:  # inaccurate/primal infeasible
            primalStatus   = SS_INFEASIBLE
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_INFEASIBLE
        elif statusCode == 2:  # dual infeasible
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_INFEASIBLE
            problemStatus  = PS_UNBOUNDED
        elif statusCode == 12:  # inaccurate/dual infeasible
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_INFEASIBLE
            problemStatus  = PS_UNBOUNDED
        elif statusCode == -1:  # iteration limit reached
            primalStatus   = SS_PREMATURE
            dualStatus     = SS_PREMATURE
            problemStatus  = PS_UNKNOWN
        elif statusCode == -2:  # search direction unreliable
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_UNSTABLE
        elif statusCode == -3:  # numerically problematic
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_UNSTABLE
        elif statusCode == -4:  # interrupted
            primalStatus   = SS_PREMATURE
            dualStatus     = SS_PREMATURE
            problemStatus  = PS_UNKNOWN
        elif statusCode == -7:  # solver error
            primalStatus   = SS_FAILURE
            dualStatus     = SS_FAILURE
            problemStatus  = PS_UNKNOWN
        else:
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_UNKNOWN

        return self._make_solution(primals, duals, primalStatus, dualStatus,
            problemStatus, {"ecos_sol": solution})


# --------------------------------------
__all__ = api_end(_API_START, globals())



#!/usr/bin/python

#-------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS Release Scripts.
#
# PICOS Release Scripts are free software: you can redistribute it and/or modify
# them under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS Release Scripts are distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-------------------------------------------------------------------------------

import os

try:
    from setuptools import setup
    from setuptools.command.sdist import sdist as _sdist
except ImportError:
    from distutils.core import setup
    from distutils.command.sdist import sdist as _sdist
    if hasattr(os, 'link'):  # prevent distutils hardlinking
        del os.link

try:
    from version import get_version, get_base_version
except ImportError:
    # PyPI strips version.py from the source package as it's not part of the installation.
    LOCATION = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    VERSION_FILE = os.path.join(LOCATION, "picos", ".version")

    def get_base_version():
        with open(VERSION_FILE, "r") as versionFile:
            return versionFile.read().strip()

    get_version = get_base_version


class sdist(_sdist):
    """sdist command, changing .version to `get_version()` result."""
    def make_release_tree(self, base_dir, files):
        """Call `super().make_relese_tree()` and change .version"""
        _sdist.make_release_tree(self, base_dir, files)
        with open(os.path.join(base_dir, "picos", ".version"), 'w') as file:
            file.write(get_version() + '\n')

setup(
    name = 'PICOS',
    version = get_version(),
    author = 'G. Sagnol, M. Stahlberg',
    author_email = 'incoming+picos-api/picos@incoming.gitlab.com',
    packages = [
        'picos',
        'picos.constraints',
        'picos.expressions',
        'picos.modeling',
        'picos.reforms',
        'picos.solvers'
    ],
    package_data = {'picos': ['.version']},
    description = 'A Python interface to conic optimization solvers.',
    long_description = open('README.rst', 'rb').read().decode('utf8'),
    long_description_content_type = 'text/x-rst',
    install_requires = [
        "CVXOPT >= 1.1.4",
        "numpy  >= 1.6.2"
    ],
    keywords = [
        'conic optimization',
        'convex optimization'
        'linear programming',
        'quadratic programming',
        'second order cone programming',
        'semidefinite programming',
        'exponential cone programming'
    ],
    classifiers = [
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
    ],
    url = 'https://gitlab.com/picos-api/picos',
    download_url = 'https://gitlab.com/picos-api/picos/tags/v{}'.format(get_base_version()),
    cmdclass = {'sdist': sdist}
)



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2019 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`PowerTraceConstraint`."""

from __future__ import division

import math
from collections import namedtuple

import cvxopt as cvx

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class PowerTraceConstraint(Constraint):
    """Bound on the trace over the :math:`p`-th power of a matrix.

    For scalar expressions, this is simply a bound on their :math:`p`-th power.
    """

    class Conversion(ConstraintConversion):
        """Bound on the :math:`p`-th power of a trace constraint conversion.

        The conversion is based on
        `this paper <http://nbn-resolving.de/urn:nbn:de:0297-zib-17511>`_.
        """

        @classmethod
        def _count_number_tree_node_types(cls, x):
            """Count number of conversion tree nodes.

            Consider a binary tree with x[i] leaves of type i, arranged from
            left to right, with sum(x) a power of two. A node of the tree is of
            type i if its 2 parents are of type i; otherwise, a new type is
            created for this node. This function counts the number of additional
            types we need to create while growing the tree.
            """
            x = [xi for xi in x if xi != 0]
            sum_x = sum(x)

            # We have reached the tree root. Stop the recursion.
            if sum_x == 1:
                return 0

            # Make sure x is a power of two.
            _log2_sum_x = math.log(sum_x, 2)
            assert _log2_sum_x == int(_log2_sum_x)

            new_x  = []
            new_t  = 0
            s      = 0
            offset = 0

            # Compute the vector new_x of types at next level.
            for x_i in x:
                s += x_i

                if s % 2 == 0:
                    if x_i - offset >= 2:
                        new_x.append((x_i - offset) // 2)

                    offset = 0
                else:
                    if   x_i - offset >= 2:
                        new_x.extend([(x_i - offset) // 2, 1])
                    elif x_i - offset == 1:
                        new_x.append(1)
                    elif x_i - offset == 0:
                        assert False, "Unexpected case."

                    offset = 1
                    new_t += 1

            assert 2*sum(new_x) == sum_x

            return new_t + cls._count_number_tree_node_types(new_x)

        @staticmethod
        def _np2(n):
            """Compute the smallest power of two that is an upper bound."""
            return 2**int(math.ceil(math.log(n, 2)))

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import (HermitianVariable, RealVariable,
                                       SymmetricVariable)
            from . import (AffineConstraint, ComplexLMIConstraint,
                           RSOCConstraint, LMIConstraint)

            n, num, den, hasM, complex = subtype

            if num > den > 0:
                x = [den, cls._np2(num) - num, num - den]
            elif num / den < 0:
                num = abs(num)
                den = abs(den)
                x = [den, num, cls._np2(num + den) - num - den]
            elif 0 < num < den:
                x = [num, cls._np2(den) - den, den - num]
            else:
                assert False, "Unexpected exponent."

            N = cls._count_number_tree_node_types(x)

            if n == 1:
                yield ("var", RealVariable.make_var_type(dim=1, bnd=0), N - 1)
                yield ("con", RSOCConstraint.make_type(argdim=1), N)
                if hasM:
                    yield ("con",
                        AffineConstraint.make_type(dim=1, eq=False), 1)
            else:
                if complex:
                    yield ("var",
                        HermitianVariable.make_var_type(dim=n**2, bnd=0), N)
                    yield ("con",
                        ComplexLMIConstraint.make_type(diag=2*n), N)
                else:
                    yield ("var", SymmetricVariable.make_var_type(
                        dim=(n * (n + 1)) // 2, bnd=0), N)
                    yield ("con", LMIConstraint.make_type(diag=2*n), N)
                yield ("con", AffineConstraint.make_type(dim=1, eq=False), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import Constant
            from ..expressions.algebra import rsoc
            from ..modeling import Problem

            x   = con.power.x
            n   = con.power.n
            num = con.power.num
            den = con.power.den
            rhs = con.rhs
            m   = con.power.m

            vtype = "hermitian" if x.complex else "symmetric"

            P = Problem()

            if n == 1:
                idt = Constant('1', 1)
                varcnt = 0
                v = []
            else:
                idt = Constant('I', cvx.spdiag([1.] * n))
                varcnt = 1
                v = [P.add_variable('__v[0]', (n, n), vtype)]

            if con.relation == Constraint.LE and num > den:
                pown = cls._np2(num)

                if n == 1:
                    lis = [rhs]*den  + [x]*(pown - num) + [idt]*(num - den)
                else:
                    lis = [v[0]]*den + [x]*(pown - num) + [idt]*(num - den)

                while len(lis) > 2:
                    newlis = []
                    while lis:
                        v1 = lis.pop()
                        v2 = lis.pop()

                        if v1 is v2:
                            newlis.append(v2)
                        else:
                            if n == 1:
                                v0 = P.add_variable(
                                    '__v[' + str(varcnt) + ']', 1)
                                P.add_constraint((v1 & v2 & v0) << rsoc())
                            else:
                                v0 = P.add_variable(
                                    '__v[' + str(varcnt) + ']', (n, n), vtype)
                                P.add_constraint(((v1 & v0) // (v0 & v2)) >> 0)

                            varcnt += 1
                            newlis.append(v0)
                            v.append(v0)
                    lis = newlis

                if n == 1:
                    P.add_constraint((lis[0] & lis[1] & x) << rsoc())
                else:
                    P.add_constraint(
                        ((lis[0] & x) // (x & lis[1])) >> 0)
                    P.add_constraint((idt | v[0]) <= rhs)
            elif con.relation == Constraint.LE and num <= den:
                num = abs(num)
                den = abs(den)

                pown = cls._np2(num + den)

                if n == 1:
                    lis = [rhs] * den + [x] * num + [idt] * (pown - num - den)
                else:
                    lis = [v[0]] * den + [x] * num + [idt] * (pown - num - den)

                while len(lis) > 2:
                    newlis = []
                    while lis:
                        v1 = lis.pop()
                        v2 = lis.pop()

                        if v1 is v2:
                            newlis.append(v2)
                        else:
                            if n == 1:
                                v0 = P.add_variable(
                                    '__v[' + str(varcnt) + ']', 1)
                                P.add_constraint((v1 & v2 & v0) << rsoc())
                            else:
                                v0 = P.add_variable(
                                    '__v[' + str(varcnt) + ']', (n, n), vtype)
                                P.add_constraint(((v1 & v0) // (v0 & v2)) >> 0)

                            varcnt += 1
                            newlis.append(v0)
                            v.append(v0)
                    lis = newlis

                if n == 1:
                    P.add_constraint((lis[0] & lis[1] & 1) << rsoc())
                else:
                    P.add_constraint(((lis[0] & idt) // (idt & lis[1])) >> 0)
                    P.add_constraint((idt | v[0]) <= rhs)
            elif con.relation == Constraint.GE:
                pown = cls._np2(den)

                if n == 1:
                    lis = [x]*num + [rhs]*(pown - den)  + [idt]*(den - num)

                else:
                    lis = [x]*num + [v[0]]*(pown - den) + [idt]*(den - num)

                while len(lis) > 2:
                    newlis = []
                    while lis:
                        v1 = lis.pop()
                        v2 = lis.pop()

                        if v1 is v2:
                            newlis.append(v2)
                        else:
                            if n == 1:
                                v0 = P.add_variable(
                                    '__v[' + str(varcnt) + ']', 1)
                                P.add_constraint((v1 & v2 & v0) << rsoc())
                            else:
                                v0 = P.add_variable(
                                    '__v[' + str(varcnt) + ']', (n, n), vtype)
                                P.add_constraint(((v1 & v0) // (v0 & v2)) >> 0)

                            varcnt += 1
                            newlis.append(v0)
                            v.append(v0)
                    lis = newlis

                if n == 1:
                    if m is None:
                        P.add_constraint((lis[0] & lis[1] & rhs) << rsoc())
                    else:
                        P.add_constraint((lis[0] & lis[1] & v[0]) << rsoc())
                        P.add_constraint((m * v[0]) > rhs)
                else:
                    P.add_constraint(((lis[0] & v[0]) // (v[0] & lis[1])) >> 0)
                    if m is None:
                        P.add_constraint((idt | v[0]) > rhs)
                    else:
                        P.add_constraint((m | v[0]) > rhs)
            else:
                assert False, "Dijkstra-IF fallthrough."

            return P

    def __init__(self, power, relation, rhs):
        """Construct a :class:`PowerTraceConstraint`.

        :param ~picos.expressions.PowerTrace ower:
            Left hand side expression.
        :param str relation:
            Constraint relation symbol.
        :param ~picos.expressions.AffineExpression rhs:
            Right hand side expression.
        """
        from ..expressions import AffineExpression, PowerTrace

        assert isinstance(power, PowerTrace)
        assert relation in self.LE + self.GE
        assert isinstance(rhs, AffineExpression)
        assert len(rhs) == 1

        p = power.p

        assert p != 0 and p != 1, \
            "The PowerTraceConstraint should not be created for p = 0 and " \
            "p = 1 as there are more direct ways to represent such powers."

        if relation == self.LE:
            assert p <= 0 or p >= 1, \
                "Upper bounding p-th power needs p s.t. the power is convex."
        else:
            assert p >= 0 and p <= 1, \
                "Lower bounding p-th power needs p s.t. the power is concave."

        self.power    = power
        self.relation = relation
        self.rhs      = rhs

        super(PowerTraceConstraint, self).__init__(power._typeStr)

    # HACK: Support Constraint's LHS/RHS interface.
    # TODO: Add a unified interface for such constraints?
    lhs = property(lambda self: self.power)

    def is_trace(self):
        """Whether the bound concerns a trace as opposed to a scalar."""
        return self.power.n > 1

    Subtype = namedtuple("Subtype", ("diag", "num", "den", "hasM", "complex"))

    def _subtype(self):
        return self.Subtype(*self.power.subtype)

    @classmethod
    def _cost(cls, subtype):
        n = subtype.diag
        if subtype.complex:
            return n**2 + 1
        else:
            return n*(n + 1)//2 + 1

    def _expression_names(self):
        yield "power"
        yield "rhs"

    def _str(self):
        if self.relation == self.LE:
            return glyphs.le(self.power.string, self.rhs.string)
        else:
            return glyphs.ge(self.power.string, self.rhs.string)

    def _get_slack(self):
        if self.relation == self.LE:
            return self.rhs.value - self.power.value
        else:
            return self.power.value - self.rhs.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Optimization problem solution representation."""

import warnings

from .. import glyphs
from ..apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


# Solution status strings, as verified by PICOS.
VS_UNKNOWN           = "unverified"
"""PICOS failed to verify the solution."""

VS_DETACHED          = "detached"
"""The solution is not attached to a problem (it was given by the user)."""

VS_EMPTY             = "empty"
"""The solution is empty; there are neither primals nor duals."""

VS_DETACHED_EMPTY    = "detached empty"
"""The solution is both detached and empty."""

VS_OUTDATED          = "outdated"
"""The solution does not fit the problem formulation any more.

Variables or constraints were removed from the problem."""

VS_INCOMPLETE        = "incomplete"
"""The primal (dual) solution does not concern all variables (constraints)."""

VS_FEASIBLE          = "feasible"
"""The solution is primal feasible; there is no dual solution."""

VS_INFEASIBLE        = "infeasible"
"""The solution is primal infeasible; there is no dual solution."""

VS_PRIMAL_FEASIBLE   = "primal feasible"
"""The solution is primal feasible; a dual solution was not verified."""

VS_PRIMAL_INFEASIBLE = "primal infeasible"
"""The solution is primal infeasible; a dual solution was not verified."""


# Primal or dual solution (or search) status strings, as claimed by the solver.
SS_UNKNOWN           = "unknown"
"""The solver did not make a clear claim about the solution status."""

SS_EMPTY             = "empty"
"""The solver claims not to have produced a solution."""

SS_OPTIMAL           = "optimal"
"""The solution is optimal."""

SS_FEASIBLE          = "feasible"
"""The solution is feasible."""

SS_INFEASIBLE        = "infeasible"
"""No feasible solution exists.

In the case of a primal solution, the problem is infeasible. In the case of a
dual solution, the problem is unbounded.
"""

SS_PREMATURE         = "premature"
"""The search was prematurely terminated due to some limit."""

SS_FAILURE           = "failure"
"""The search was termined due to a solver failure."""


# Problem status strings, as claimed by the solver.
PS_UNKNOWN           = "unknown"
"""The solver did not make a clear claim about the problem status."""

PS_FEASIBLE          = "feasible"
"""The problem is primal (and dual) feasible and bounded."""

PS_INFEASIBLE        = "infeasible"
"""The problem is primal infeasible (and dual unbounded or infeasible)."""

PS_UNBOUNDED         = "unbounded"
"""The problem is primal unbounded (and dual infeasible)."""

PS_INF_OR_UNB        = "infeasible or unbounded"
"""The problem is primal infeasible or unbounded.

Being unbounded is usually infered from being dual infeasible."""

PS_UNSTABLE          = "unstable"
"""The problem was found numerically unstable or otherwise hard to handle."""

PS_ILLPOSED          = "illposed"
"""The problem was found to be in a state that is not amenable to solution."""


def _check_type(argument, *types):
    """Enforce the type of a method or function argument."""
    for type_ in types:
        if type_ is None:
            type_ = type(None)
        if isinstance(argument, type_):
            return

    raise TypeError("An argument is of type '{}' but must be instance of {}."
        .format(type(argument).__name__, " or ".join("'{}'".format(t.__name__)
            for t in types)))


class Solution:
    """Assignment of primal and dual values to variables and constraints.

    Instances are usually returned by a solver (and thus bound to a
    :class:`problem <picos.Problem>` instance), but may be manually created by
    the user:

    >>> import picos
    >>> P = picos.Problem()
    >>> x = P.add_variable("x")
    >>> s = picos.Solution({x: 1}); s
    <detached primal solution from user>
    >>> s.apply()
    >>> x.value
    1.0

    If the solution was created by a solver (or attached to a problem via
    :func:`attach_to`), more information is available:

    >>> C1 = P.add_constraint(x >= 2)
    >>> s = P.minimize(x, solver = "cvxopt", duals = False); s
    <feasible primal solution (claimed optimal) from cvxopt>
    >>> "{:.2f} ms".format(1000.0 * s.searchTime) #doctest: +SKIP
    '0.83 ms'
    >>> C2 = P.add_constraint(x >= 3); s
    <infeasible primal solution (was feasible and claimed optimal) from cvxopt>
    """

    def __init__(self, primals, duals=None, problem=None, solver="user",
            primalStatus=SS_UNKNOWN, dualStatus=SS_UNKNOWN,
            problemStatus=PS_UNKNOWN, searchTime=0.0, info=None,
            vectorizedPrimals=False):
        """Create a solution to an optimization problem.

        :param dict(picos.expressions.BaseVariable, object) primals:
            A mapping of variables to their primal solution value.
        :param dict(picos.constraints.Constraint, object) duals:
            A mapping of constraints to their dual solution value.
        :param picos.Problem problem:
            The problem that was solved to create the solution. If ``None``,
            then the solution is "detached".
        :param str solver:
            The name of the solver that was used to create the solution.
        :param str primalStatus:
            The primal solution status as reported by the solver.
        :param str dualStatus:
            The dual solution status as reported by the solver.
        :param str problemStatus:
            The state of the problem as reported by the solver.
        :param float searchTime:
            Seconds that the solution process took.
        :param dict info:
            Additional solution (meta)data.
        :param bool vectorizedPrimals:
            Whether primal solution values are given with respect to the
            variable's special vectorization format as used by PICOS internally.
        """
        from ..expressions import BaseVariable
        from ..constraints import Constraint
        from .problem      import Problem

        if duals is None:
            duals = {}

        if info is None:
            info = {}

        # Be strict about the arguments as they are handed to the user.
        _check_type(primals, dict)
        _check_type(problem, None, Problem)
        _check_type(duals, dict)
        _check_type(solver, str)
        _check_type(primalStatus, str)
        _check_type(dualStatus, str)
        _check_type(problemStatus, str)
        _check_type(searchTime, float)
        _check_type(info, dict)
        _check_type(vectorizedPrimals, bool)

        for variable, _ in primals.items():
            if not isinstance(variable, BaseVariable):
                raise TypeError("They keys in the primals argument of "
                    "Solution.__init__ must be variables.")

        for constraint, _ in duals.items():
            if not isinstance(constraint, Constraint):
                raise TypeError("They keys in the duals argument of "
                    "Solution.__init__ must be constraints.")

        # Derive a "claimed status" from the claimed primal and dual states.
        if primals and duals:
            if primalStatus == dualStatus:
                claimedStatus = primalStatus
            else:
                claimedStatus = "primal {} and dual {}".format(
                    primalStatus, dualStatus)
        elif primals:
            # Do not warn about correctingdualStatus, because the solver might
            # have produced primals but PICOS did not read them.
            dualStatus    = SS_EMPTY
            claimedStatus = primalStatus
        elif duals:
            # Do not warn about correcting primalStatus, because the solver
            # might have produced duals but PICOS did not read them.
            primalStatus  = SS_EMPTY
            claimedStatus = dualStatus
        else:
            primalStatus  = SS_EMPTY
            dualStatus    = SS_EMPTY
            claimedStatus = SS_EMPTY

        # Infeasible problem implies infeasible primal.
        if problemStatus == PS_INFEASIBLE \
        and primalStatus not in (SS_INFEASIBLE, SS_EMPTY):
            warnings.warn(
                "{} claims that a problem is infeasible but does not say the "
                "same about the nonempty primal solution. Correcting this.".
                format(solver), RuntimeWarning)
            primalStatus = SS_INFEASIBLE

        # Unbounded problem implies infeasible dual.
        if problemStatus == PS_UNBOUNDED \
        and dualStatus not in (SS_INFEASIBLE, SS_EMPTY):
            warnings.warn(
                "{} claims that a problem is unbounded but does not say that "
                "the nonempty dual solution is infeasible. Correcting this.".
                format(solver), RuntimeWarning)
            dualStatus = SS_INFEASIBLE

        # Optimal solution implies feasible problem.
        if claimedStatus == SS_OPTIMAL and problemStatus != PS_FEASIBLE:
            warnings.warn(
                "{} claims to have found an optimal solution but does not say "
                " that the problem is feasible. Correcting this."
                .format(solver), RuntimeWarning)
            problemStatus = PS_FEASIBLE

        self.problem       = problem
        """The problem that was solved to produce the solution."""

        self.solver        = solver
        """The solver that produced the solution."""

        self.searchTime    = searchTime
        """Time in seconds that the solution search took."""

        self.primals       = primals
        """The primal solution values returned by the solver."""

        self.duals         = duals
        """The dual solution values returned by the solver."""

        self.info          = info
        """Additional information provided by the solver."""

        self.lastStatus    = VS_UNKNOWN
        """The solution status as verified by PICOS when the solution was
        applied to the problem."""

        self.primalStatus  = primalStatus
        """The primal solution status as claimed by the solver."""

        self.dualStatus    = dualStatus
        """The dual solution status as claimed by the solver."""

        self.claimedStatus = claimedStatus
        """The primal and dual solution status as claimed by the solver."""

        self.problemStatus = problemStatus
        """The problem status as claimed by the solver."""

        self.vectorizedPrimals = vectorizedPrimals
        """Whether primal values refer to variables' special vectorizations."""

    def _status_of_problem(self, problem):
        """Retrieve the problem's verified solution status.

        Requires that the solution has just been applied to the problem.
        """
        if not self.primals and not self.duals:
            return VS_EMPTY

        try:
            isFeasible = problem.check_current_value_feasibility()[0]
        except LookupError:
            return VS_INCOMPLETE
        except Exception:
            return VS_UNKNOWN

        if isFeasible:
            return VS_PRIMAL_FEASIBLE if self.duals else VS_FEASIBLE
        else:
            return VS_PRIMAL_INFEASIBLE if self.duals else VS_INFEASIBLE

    @property
    def status(self):
        """The current solution status as verified by PICOS.

        .. warning::

            Accessing this attribute is expensive for large problems as a copy
            of the problem needs to be created and valued. If you have just
            applied the solution to a :class:`problem <picos.Problem>`, query
            the solution's lastStatus attribute instead.
        """
        if not self.primals and not self.duals:
            if not self.problem:
                return VS_DETACHED_EMPTY
            else:
                return VS_EMPTY
        elif not self.problem:
            return VS_DETACHED
        elif not self.primals:
            return VS_UNKNOWN

        problemCopy = self.problem.copy()

        try:
            self.apply(toProblem=problemCopy)
        except RuntimeError:
            return VS_OUTDATED

        return self._status_of_problem(problemCopy)

    @property
    def value(self):
        """The objective value of the solution.

        .. warning::

            Accessing this attribute is expensive for large problems as a copy
            of the problem needs to be created and valued. If you have just
            applied the solution to a :class:`problem <picos.Problem>`, query
            that problem instead.
        """
        if not self.problem:
            raise RuntimeError(
                "Cannot compute the objective value of a detached solution. "
                "Use attach_to to assign the solution to a problem.")

        problemCopy = self.problem.copy()

        self.apply(toProblem=problemCopy)

        return problemCopy.value

    def __str__(self):
        verifiedStatus = self.status
        lastStatus     = self.lastStatus
        claimedStatus  = self.claimedStatus
        problemStatus  = self.problemStatus

        if self.primals and self.duals:
            solutionType = "solution pair"
        elif self.primals:
            solutionType = "primal solution"
        elif self.duals:
            solutionType = "dual solution"
        else:
            solutionType = "solution"  # "(detached) empty solution"

        # Print the last status if it is known and differs from the current one.
        printLastStatus = lastStatus != VS_UNKNOWN and \
            verifiedStatus != lastStatus

        # Print the claimed status only if it differs from the initial verified
        # one is not implied by a problem status that will be printed.
        printClaimedStatus = \
            claimedStatus not in (verifiedStatus, SS_UNKNOWN) and \
            problemStatus not in (PS_INFEASIBLE, PS_UNBOUNDED)

        # Print the problem status only if it is interesting.
        printProblemStatus = \
            problemStatus not in (PS_UNKNOWN, PS_FEASIBLE)

        if printLastStatus and printClaimedStatus:
            unverifiedStatus = " (was {} and claimed {})".format(
                lastStatus, claimedStatus)
        elif printLastStatus:
            unverifiedStatus = " (was {})".format(lastStatus)
        elif printClaimedStatus:
            unverifiedStatus = " (claimed {})".format(claimedStatus)
        else:
            unverifiedStatus = ""

        if printProblemStatus:
            unverifiedStatus += \
                " for a problem claimed {}".format(problemStatus)

        return "{} {}{} from {}".format(verifiedStatus, solutionType,
            unverifiedStatus, self.solver)

    def __repr__(self):
        return glyphs.repr1(self.__str__())

    def apply(self, primals=True, duals=True, clearOnNone=True, toProblem=None,
              snapshotStatus=False):
        """Apply the solution to the involved variables and constraints.

        :param bool primals: Whether to apply the primal solution.
        :param bool duals: Whether to apply the dual solution.
        :param bool clearOnNone: Whether to clear the value of a variable or
            constraint if the solution has it set to None. This could happen in
            case of an error or shortcoming of the solver or PICOS.
        :param picos.Problem toProblem: If set to a copy of the problem that was
            used to produce the solution, will apply the solution to that copy's
            variables and constraints instead.
        :param bool snapshotStatus: Whether to update the lastStatus attribute
            with the new (verified) solution status. PICOS enables this whenever
            it applies a solution returned by a solver.
        """
        if toProblem:
            if primals:
                thePrimals = {}
                try:
                    for variable, primal in self.primals.items():
                        thePrimals[toProblem.variables[variable.name]] = primal
                except KeyError:
                    raise RuntimeError(
                        "Cannot apply solution to specified problem as not all "
                        "variables for which primal values exist were found.")

            if duals:
                theDuals = {}
                try:
                    for constraint, dual in self.duals.items():
                        theDuals[toProblem.constraints[constraint.id]] = dual
                except KeyError:
                    raise RuntimeError(
                        "Cannot apply solution to specified problem as not all "
                        "constraints for which dual values exist were found.")
        else:
            thePrimals = self.primals
            theDuals   = self.duals

        if primals:
            for variable, primal in thePrimals.items():
                if primal is None and not clearOnNone:
                    continue

                if self.vectorizedPrimals:
                    variable.internal_value = primal
                else:
                    variable.value = primal

        if duals:
            for constraint, dual in theDuals.items():
                if dual is None and not clearOnNone:
                    continue
                constraint.dual = dual

        if snapshotStatus:
            if toProblem:
                self.lastStatus = self._status_of_problem(toProblem)
            elif self.problem:
                self.lastStatus = self._status_of_problem(self.problem)
            else:  # detached solution
                self.lastStatus = self.status

        if toProblem:
            toProblem._last_solution = self
        elif self.problem:
            self.problem._last_solution = self

    def attach_to(self, problem, snapshotStatus=False):
        """Attach (or move) the solution to a problem.

        Only variables and constraints that exist on the problem (same name or
        ID, respectively) are kept.

        :param bool snapshotStatus: Whether to set the lastStatus attribute
            of the copy to match the new problem.
        """
        self.problem = problem

        # Find variables of same name in the problem and assign primals.
        oldPrimals, self.primals = self.primals, {}
        for variable, primal in oldPrimals.items():
            if variable.name in problem.variables:
                self.primals[problem.variables[variable.name]] = primal

        # Find constraints of same ID in the problem and assign duals.
        oldDuals, self.duals = self.duals, {}
        for constraint, dual in oldDuals.items():
            if constraint.id in problem.constraints:
                self.duals[problem.constraints[constraint.id]] = dual

        # Update the last (verified) status.
        if snapshotStatus:
            self.lastStatus = problem.status
        else:
            self.lastStatus = VS_UNKNOWN


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Quadratic constraint types."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class NonconvexQuadraticConstraint(Constraint):
    """Bound on a nonconvex quadratic expression."""

    def __init__(self, lhs, relation, rhs):
        """Construct a :class:`NonconvexQuadraticConstraint`.

        :param lhs: Left hand side quadratic or affine expression.
        :type lhs: ~picos.expressions.QuadraticExpression or
            ~picos.expressions.AffineExpression
        :param str relation: Constraint relation symbol.
        :param rhs: Right hand side quadratic or affine expression.
        :type rhs: ~picos.expressions.QuadraticExpression or
            ~picos.expressions.AffineExpression
        """
        from ..expressions import AffineExpression, QuadraticExpression

        assert isinstance(lhs, (AffineExpression, QuadraticExpression))
        assert isinstance(rhs, (AffineExpression, QuadraticExpression))
        assert any(isinstance(exp, QuadraticExpression) for exp in (lhs, rhs))
        assert relation in self.LE + self.GE
        assert lhs.size == rhs.size

        self.lhs      = lhs
        self.rhs      = rhs
        self.relation = relation

        Constraint.__init__(self, self._get_type_term())

    def _get_type_term(self):
        return "Nonconvex Quadratic"

    # TODO: Add common interface for all the le0, ge0, smaller, greater methods?
    @cached_property
    def le0(self):
        """Quadratic expression constrained to be at most zero."""
        if self.relation == self.GE:
            return self.rhs - self.lhs
        else:
            return self.lhs - self.rhs

    @property
    def smaller(self):
        """Smaller-or-equal side of the constraint."""
        return self.rhs if self.relation == self.GE else self.lhs

    @property
    def greater(self):
        """Greater-or-equal side of the constraint."""
        return self.lhs if self.relation == self.GE else self.rhs

    Subtype = namedtuple("Subtype", ("qterms"))

    def _subtype(self):
        return self.Subtype(self.le0.num_quad_terms)

    @classmethod
    def _cost(cls, subtype):
        return subtype.qterms

    def _expression_names(self):
        yield "lhs"
        yield "rhs"

    def _str(self):
        if self.relation == self.LE:
            return glyphs.le(self.lhs.string, self.rhs.string)
        else:
            return glyphs.ge(self.lhs.string, self.rhs.string)

    def _get_size(self):
        return (1, 1)

    def _get_slack(self):
        if self.relation == self.LE:
            return self.rhs.value - self.lhs.value
        else:
            return self.lhs.value - self.rhs.value


class ConicQuadraticConstraint(NonconvexQuadraticConstraint):
    r"""Bound on a *nearly* convex quadratic expression.

    Nearly convex means that the bilinear form representing the quadratic part
    of the expression that the constraint poses to be at most zero has exactly
    one negative eigenvalue. More precisely, if the constraint is of the form
    :math:`x^TQx + a^Tx + b \leq x^TRx + b^Tx + c`, then `Q - R` needs to have
    exactly one negative eigenvalue for the constraint to be nearly convex. Such
    constraints can be posed as conic constraints under an additional assumption
    that some affine term is nonnegative.

    At this point, this class may only be used for nonconvex constraints of the
    form :math:`x^TQx + p^Tx + q \leq (a^Tx + b)(c^Tx + d)` with
    :math:`x^TQx + p^Tx + q` representable as a squared norm. In this case, the
    additional assumptions required for a conic reformulation are
    :math:`a^Tx + b \geq 0` and :math:`b^Tx + c \geq 0`.

    Whether a constraint of this type is strengthened to a conic constraint or
    relabeled as a :class:`NonconvexQuadraticConstraint` depends on the
    :ref:`assume_conic <option_assume_conic>` option.

    :Example:

    >>> from picos import Options, RealVariable
    >>> x, y, z = RealVariable("x"), RealVariable("y"), RealVariable("z")
    >>> C = x**2 + 1 <= y*z; C
    <Conic Quadratic Constraint: x² + 1 ≤ y·z>
    >>> P = C.__class__.Conversion.convert(C, Options(assume_conic=True))
    >>> list(P.constraints.values())[0]
    <4×1 RSOC Constraint: ‖fullroot(x² + 1)‖² ≤ y·z ∧ y ≥ 0>
    >>> Q = C.__class__.Conversion.convert(C, Options(assume_conic=False))
    >>> list(Q.constraints.values())[0]
    <Nonconvex Quadratic Constraint: x² + 1 ≤ y·z>

    .. note::

        Solver implementations must not support this constraint type so that
        the user's choice for the :ref:`assume_conic <option_assume_conic>`
        option is respected.
    """

    class Conversion(ConstraintConversion):
        """Nearly convex quadratic to (rotated) second order cone conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from . import AbsoluteValueConstraint, SOCConstraint, RSOCConstraint

            if options.assume_conic:
                n = subtype.conic_argdim

                if subtype.rotated:
                    yield ("con", RSOCConstraint.make_type(argdim=n), 1)
                elif n == 1:
                    yield ("con", AbsoluteValueConstraint.make_type(), 1)
                else:
                    yield ("con", SOCConstraint.make_type(argdim=n), 1)
            else:
                yield ("con", NonconvexQuadraticConstraint.make_type(
                    qterms=subtype.qterms), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import AffineExpression
            from ..modeling import Problem
            from . import AbsoluteValueConstraint, SOCConstraint, RSOCConstraint

            P = Problem()

            if options.assume_conic:
                q, r = con.smaller, con.greater
                root = AffineExpression.zero() if q.is0 else q.fullroot

                if r.scalar_factors[0] is not r.scalar_factors[1]:
                    P.add_constraint(RSOCConstraint(root, *r.scalar_factors))
                else:
                    bound = r.scalar_factors[0]

                    if len(root) == 1:
                        P.add_constraint(AbsoluteValueConstraint(root, bound))
                    else:
                        P.add_constraint(SOCConstraint(root, bound))
            else:
                P.add_constraint(NonconvexQuadraticConstraint(
                    con.lhs, con.relation, con.rhs))

            return P

    def __init__(self, lhs, relation, rhs):
        """Construct a :class:`ConicQuadraticConstraint`.

        See :meth:`NonconvexQuadraticConstraint.__init__` for more.
        """
        from ..expressions import QuadraticExpression

        NonconvexQuadraticConstraint.__init__(self, lhs, relation, rhs)

        # Additional usage restrictions for the near-convex case.
        assert isinstance(self.lhs, QuadraticExpression)
        assert isinstance(self.rhs, QuadraticExpression)
        assert self.smaller.is_squared_norm
        assert self.greater.scalar_factors

    def _get_type_term(self):
        return "Conic Quadratic"

    Subtype = namedtuple("Subtype", ("qterms", "conic_argdim", "rotated"))

    def _subtype(self):
        qterms = self.le0.num_quad_terms
        conic_argdim = 1 if self.smaller.is0 else len(self.smaller.fullroot)
        sf = self.greater.scalar_factors
        rotated = sf[0] is not sf[1]

        return self.Subtype(qterms, conic_argdim, rotated)


class ConvexQuadraticConstraint(NonconvexQuadraticConstraint):
    """Bound on a convex quadratic expression."""

    class ConicConversion(ConstraintConversion):
        """Convex quadratic to (rotated) second order cone conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from . import AbsoluteValueConstraint, SOCConstraint, RSOCConstraint

            if subtype.haslin:
                yield ("con", RSOCConstraint.make_type(argdim=subtype.rank), 1)
            elif subtype.rank == 1:
                yield ("con", AbsoluteValueConstraint.make_type(), 1)
            else:
                yield ("con", SOCConstraint.make_type(argdim=subtype.rank), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import AffineExpression
            from ..modeling import Problem
            from . import AbsoluteValueConstraint, SOCConstraint, RSOCConstraint

            le0 = con.le0

            P = Problem()

            if con.subtype.haslin:
                one = AffineExpression.from_constant(1)

                P.add_constraint(RSOCConstraint(le0.quadroot, -le0._aff, one))
            else:
                assert le0._aff.constant
                value = -le0._aff.value

                if value < 0:
                    raise ValueError("The constraint {} is infeasible as it "
                        "upper-bounds a positive semidefinite quadratic form by"
                        " a negative constant.".format(con))

                root  = le0.quadroot
                bound = AffineExpression.from_constant(value**0.5,
                    name=glyphs.sqrt(glyphs.scalar(value)))

                if len(root) == 1:
                    P.add_constraint(AbsoluteValueConstraint(root, bound))
                else:
                    P.add_constraint(SOCConstraint(root, bound))

            return P

    def __init__(self, lhs, relation, rhs):
        """Construct a :class:`ConvexQuadraticConstraint`.

        See :meth:`NonconvexQuadraticConstraint.__init__` for more.
        """
        NonconvexQuadraticConstraint.__init__(self, lhs, relation, rhs)

    def _get_type_term(self):
        return "Convex Quadratic"

    Subtype = namedtuple("Subtype", ("qterms", "rank", "haslin"))

    def _subtype(self):
        from ..expressions import QuadraticExpression

        # HACK: Be consistent with the prediction, which cannot foresee whether
        #       the linear terms of both sides cancel out.
        lhs_aff_is_cst = self.lhs._aff.constant \
            if isinstance(self.lhs, QuadraticExpression) else self.lhs.constant
        rhs_aff_is_cst = self.rhs._aff.constant \
            if isinstance(self.rhs, QuadraticExpression) else self.rhs.constant
        haslin = not lhs_aff_is_cst or not rhs_aff_is_cst

        return self.Subtype(self.le0.num_quad_terms, self.le0.rank, haslin)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`SimplexConstraint`."""

from collections import namedtuple

import cvxopt

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class SimplexConstraint(Constraint):
    """(Symmetrized, truncated) simplex membership constraint."""

    class AffineConversion(ConstraintConversion):
        """Simplex membership to affine constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable as RV
            from . import AffineConstraint as AC

            n           = subtype.argdim
            truncated   = subtype.truncated
            symmetrized = subtype.symmetrized

            if not truncated and not symmetrized:
                yield ("con", AC.make_type(dim=(n + 1), eq=False), 1)
            elif truncated and not symmetrized:
                yield ("con", AC.make_type(dim=(2*n + 1), eq=False), 1)
            elif not truncated and symmetrized:
                yield ("var", RV.make_var_type(dim=n, bnd=0), 1)
                yield ("con", AC.make_type(dim=(n + 1), eq=False), 2)
                yield ("con", AC.make_type(dim=1, eq=False), 1)
            else:
                yield ("var", RV.make_var_type(dim=n, bnd=0), 1)
                yield ("con", AC.make_type(dim=(n + 1), eq=False), 2)
                yield ("con", AC.make_type(dim=1, eq=False), 1)
                yield ("con", AC.make_type(dim=n, eq=False), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import RealVariable
            from ..modeling import Problem

            simplex     = con.simplex
            truncated   = simplex.truncated
            symmetrized = simplex.symmetrized

            x = con.element.vec
            n = len(x)
            r = simplex.radius

            P = Problem()

            if not truncated and not symmetrized:
                aff = -x // (1 | x)
                rhs = cvxopt.sparse([0]*n)
                P.add_constraint(aff <= rhs // r)
            elif truncated and not symmetrized:
                aff = x // -x // (1 | x)
                rhs = cvxopt.sparse([1]*n + [0]*n)
                P.add_constraint(aff <= rhs // r)
            elif not truncated and symmetrized:
                v = RealVariable("__v", n)
                P.add_constraint(x <= v)
                P.add_constraint(-x <= v)
                P.add_constraint((1 | v) <= r)
            else:  # truncated and symmetrized
                v = RealVariable("__v", n)
                P.add_constraint(x <= v)
                P.add_constraint(-x <= v)
                P.add_constraint((1 | v) <= r)
                P.add_constraint(v <= 1)

            return P

    def __init__(self, simplex, element):
        """Construct a :class:`SimplexConstraint`.

        :param ~picos.expressions.AffineExpression element:
            Expression in the simplex.
        """
        from ..expressions import AffineExpression, Simplex

        assert isinstance(simplex, Simplex)
        assert isinstance(element, AffineExpression)

        self.simplex = simplex
        self.element = element

        super(SimplexConstraint, self).__init__(simplex._typeStr)

    Subtype = namedtuple("Subtype", ("argdim", "truncated", "symmetrized"))

    def _subtype(self):
        return self.Subtype(
            len(self.element),
            self.simplex.truncated,
            self.simplex.symmetrized)

    @classmethod
    def _cost(cls, subtype):
        return subtype.argdim

    def _expression_names(self):
        yield "simplex"
        yield "element"

    def _str(self):
        return glyphs.element(self.element.string, self.simplex.string)

    def _get_slack(self):
        # TODO: Compute simplex constraint slack.
        raise NotImplementedError

        # FIXME: This old code can't be right for all cases.
        # return cvxopt.matrix([1 - norm(self.element, 'inf').value,
        #     self.simplex.radius - norm(self.element, 1).value])


# --------------------------------------
__all__ = api_end(_API_START, globals())



# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Backend for problem reformulation classes."""

# ------------------------------------------------------------------------------
# TODO: Make the solver base class inherit from Reformulation, in particular the
#       update related methods.
# ------------------------------------------------------------------------------

from __future__ import print_function

from ..apidoc import api_end, api_start
from ..compat import ABC, abstractmethod
from ..modeling import Footprint, Problem, Solution

_API_START = api_start(globals())
# -------------------------------


class Reformulation(ABC):
    """Base class for problem reformulations.

    Abstract base class for a reformulation from one (possibly already
    reformulated) problem form to another.
    """

    # --------------------------------------------------------------------------
    # The following section contains the abstract methods, which are exactly the
    # methods that reformulation implementations are supposed to override.
    # --------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def supports(cls, footprint):
        """Whether the reformulation affects problems with the given footprint.

        The reformulation must support every problem with such a footprint and
        the resulting problem should have a changed footprint.
        """
        pass

    @classmethod
    @abstractmethod
    def predict(cls, footprint):
        """Predict the reformulation's effect on a problem footprint.

        Given a problem footprint, returns another problem footprint that a
        problem with the former one would be reformulated to.

        This is used to predict the effects of a reformulation when planning
        a solution strategy without the cost of actually transforming a problem.
        """
        pass

    @abstractmethod
    def forward(self):
        """Perform the initial problem reformulation.

        Creates a modified copy or clone of the problem in :attr:`input` and
        stores it as :attr:`output`.

        See :func:`copy <picos.Problem.copy>` and :func:`clone
        <picos.Problem.clone>` for the differences between a copy and a clone.

        Implementations are supposed to do the necessary bookkeeping so that
        :func:`backward` can transform a solution to the new problem back to a
        solution of the original problem.
        """
        pass

    @abstractmethod
    def update(self):
        """Update a previous problem reformulation.

        Updates :attr:`output` and related bookkeeping information with respect
        to changes in :attr:`input`.

        :raises NotImplementedError:
            If performing an update is not feasible for the reformulation.
        """
        pass

    @abstractmethod
    def backward(self, solution):
        """Translate back a solution from reformulated to original problem.

        Transforms a single :class:`solution <picos.Solution>` to :attr:`output`
        to a solution of :attr:`input`.

        The method is allowed to modify the solution; it is not necessary to
        work on a copy. In particular, :func:`attach_to
        <picos.Solution.attach_to>` can be used if :func:`forward` has created
        a deep copy of the problem.
        """
        pass

    # --------------------------------------------------------------------------
    # The following section contains the non-abstract methods, which are exactly
    # the methods that reformulation implementations are supposed to inherit.
    # --------------------------------------------------------------------------

    def _reset_knowns(self):
        self._knownObjective   = None
        self._knownVariables   = set()
        self._knownConstraints = set()

    def _set_knowns(self):
        self._knownObjective   = self.input.objective
        self._knownVariables   = set(self.input.variables.values())
        self._knownConstraints = set(self.input.constraints.values())

    def __init__(self, theObject):
        """Initialize :class:`Reformulation` instances.

        :param theObject: The input to work on; either an optimization problem
            or the (future) output of another reformulation.
        :type theObject:
            ~picos.Problem or ~picos.reforms.reformulation.Reformulation
        """
        if isinstance(theObject, Problem):
            self.predecessor = None
            self._input      = theObject
        elif isinstance(theObject, Reformulation):
            self.predecessor = theObject
            theObject.successor = self
        elif isinstance(theObject, Footprint):
            raise TypeError("Reformulations cannot be instanciated using "
                "problem footprints. Use the predict classmethod.")
        else:
            raise TypeError("Cannot reformulate an object of type '{}'."
                .format(type(theObject).__name__))

        self.successor = None
        """The next reformulation in the pipeline."""

        self.output = None
        """The output problem."""

        self._reset_knowns()

    def reset(self):
        """Reset the pipeline from this reformulation onward.

        This is done whenever a reformulation does not implement :meth:`update`
        so that succeeding reformulations do not attempt to update a problem
        which was completely rewritten as this may be inefficient.
        """
        assert self.successor, \
            "The reformulation being reset has no successor."

        self.output = None
        self._reset_knowns()

        self.successor.reset()

    input = property(lambda self:
        self.predecessor.output if self.predecessor else self._input,
        doc="The input problem.")

    verbosity = property(lambda self: self.input.options.verbosity,
        doc="Verbosity level of the reformulation; same as for input problem.")

    def _verify_prediction(self):
        if not self.input.options.verify_prediction:
            return

        expectedType = self.predict(Footprint.from_problem(self.input))
        outputType = Footprint.from_problem(self.output)

        if outputType != expectedType:
            raise RuntimeError("{} failed to produce a problem with the "
                "expected footprint:\nEXPECTED: {}\nOUTCOME : {}\n"
                "This is a bug; please report it to the PICOS developers. "
                "You can disable the 'verify_prediction' option to try solving "
                "anyway.".format(type(self).__name__, expectedType, outputType))

    def execute(self):
        """Reformulate the problem and obtain a solution from the result.

        For this to work there needs to be a solver instance at the end of the
        reformulation pipeline, which would implement its own version of this
        method that actually solves the problem and produces the first solution.
        """
        assert self.successor, \
            "The reformulation being executed has no successor."

        verbose = self.verbosity > 0

        # Update the output problem if possible.
        if self.output:
            if verbose:
                print("Updating {}.".format(self.__class__.__name__))

            try:
                self.update()
            except NotImplementedError:
                if verbose:
                    print("Update failed: Not implemented for {}."
                        .format(self.__class__.__name__))

                self.reset()

        # Create the output problem if necessary.
        if not self.output:
            if verbose:
                print("Applying {}.".format(self.__class__.__name__))

            self.forward()
            self._set_knowns()

        # Verify that the output problem is of the expected type.
        self._verify_prediction()

        # Advance one step.
        outputSolution = self.successor.execute()

        # Transform the solution of the output problem for the input problem.
        if isinstance(outputSolution, Solution):
            return self.backward(outputSolution)
        else:
            return [self.backward(solution) for solution in outputSolution]

    def _objective_has_changed(self):
        """Check for an objective function change.

        :returns: Whether the optimization objective has changed since the last
            forward or update.
        """
        assert self._knownObjective is not None, \
            "_objective_has_changed may only be used inside _update_problem."

        objectiveChanged = self._knownObjective != self.input.objective

        if objectiveChanged:
            self._knownObjective = self.input.objective

        return objectiveChanged

    def _new_variables(self):
        """Check for new variables.

        Yields variables that were added to the input problem since the last
        forward or update.

        Note that variables received from this method will also be added to the
        set of known variables, so you can only iterate once within each update.
        """
        for variable in self.input.variables.values():
            if variable not in self._knownVariables:
                self._knownVariables.add(variable)
                yield variable

    def _removed_variables(self):
        """Check for removed variables.

        Yields variables that were removed from the input problem since the last
        forward or update.

        Note that variables received from this method will also be removed from
        the set of known variables, so you can only iterate once within each
        update.
        """
        newVariables = set(self.input.variables.values())
        for variable in self._knownVariables:
            if variable not in newVariables:
                yield variable
        self._knownVariables.intersection_update(newVariables)

    def _new_constraints(self):
        """Check for new constraints.

        Yields constraints that were added to the input problem since the last
        forward or update.

        Note that constraints received from this method will also be added to
        the set of known constraints, so you can only iterate once within each
        update.
        """
        for constraint in self.input.constraints.values():
            if constraint not in self._knownConstraints:
                self._knownConstraints.add(constraint)
                yield constraint

    def _removed_constraints(self):
        """Check for removed constraints.

        Yields constraints that were removed from the input problem since the
        last forward or update.

        Note that constraints received from this method will also be removed
        from the set of known constraints, so you can only iterate once within
        each update.
        """
        newConstraints = set(self.input.constraints.values())
        for constraint in self._knownConstraints:
            if constraint not in newConstraints:
                yield constraint
        self._knownConstraints.intersection_update(newConstraints)

    def _pass_updated_objective(self):
        """Pass changes in the objective function from input to output problem.

        .. warning::
            This method resets the objective-has-changed state.
        """
        if self._objective_has_changed():
            self.output.objective = self.input.objective

    # TODO: Determine if and in which form such a method is needed now that
    #       variables are added to problems implicitly (but still explicitly to
    #       solvers).
    def _pass_updated_vars(self):
        """Pass variable changes from input to output problem.

        Adds all new varibles in the input problem to the output problem, and
        removes all variables removed from the input problem also from the
        output problem.

        .. warning::
            Variables are passed as with :meth:`Problem.clone`, not copied.

        .. warning::
            This method clears the buffers of new and removed variables.
        """
        for variable in self._new_variables():
            pass

        for variable in self._removed_variables():
            pass

    def _pass_updated_cons(self, ignore=type(None)):
        """Pass constraint changes from input to output problem.

        Adds all new constraints in the input problem to the output problem, and
        removes all constraints removed from the input problem also from the
        output problem.

        :param type ignore: Constraints of this type are not handled. Instead,
            the method returns a pair `(added, removed)` that contains the
            respective constraints that were not handled.

        .. warning::
            Constraints are passed as with :meth:`Problem.clone`, not copied.

        .. warning::
            This method clears the buffers of new and removed constraints.
        """
        added, removed = [], []

        for constraint in self._new_constraints():
            assert constraint.id not in self.output.constraints
            if isinstance(constraint, ignore):
                added.append(constraint)
            else:
                self.output.add_constraint(constraint)

        for constraint in self._removed_constraints():
            if isinstance(constraint, ignore):
                # No assertion in this case, because the reformulation would not
                # have added such a constraint.
                removed.append(constraint)
            else:
                assert constraint.id in self.output.constraints
                self.output.remove_constraint(constraint.id)

        if ignore is not type(None):  # noqa: E721
            return added, removed

    def _pass_updated_options(self):
        """Make the output problem use the same options object as the input."""
        self.output.options = self.input.options


# --------------------------------------
__all__ = api_end(_API_START, globals())
